

################################################################################
#'
#' Calculates the GC content of probe sequences.
#'
#' Calculates the GC content as sum of the relative numbers of Cs and Gs in the
#' probe sequence. More precisely, the GC content is calculated by
#' \eqn{\frac{G+C}{N}} where N is the length of probe sequence while C and G are
#' the number of occurrences of the character C and G in the probe sequence,
#' respectively.
#'
#' @param  probeseq A \code{character} vector containing the probe sequences.
#' @return Returns a \code{numeric} vector containing the GC content
#'         (\eqn{0 \leq GC \leq 1}) of each probe sequence in \code{probeseq}.
#' @keywords internal
#'
TileCalcGC <- function(probeseq){
################################################################################
  stopifnot(!missing(probeseq))
  
  return(unlist(lapply(probeseq,
                       function(x){
                         a <- unlist(strsplit(x,""))
                         (length(grep("G", a)) +
                          length(grep("C", a)))/length(a)
                       })
                )
         )
}

################################################################################
#'
#' Performs full-quantile normalization.
#'
#' Performs full-quantile normalization of the given \code{numeric}
#' \code{matrix} or \code{data.frame} that contains only \code{numeric} values.
#'
#' @param  int A \code{numeric} \code{matrix} or \code{data.frame} with probe
#'         intensities to be quantile normalized. It consists of rows for each
#'         probe and columns for each sample.
#' @param  overwrite Indicates whether normalized intensities are returned
#'         by overwriting the given \code{int} parameter or by returning the
#'         normalized intensities as a \code{numeric} \code{matrix}.
#' @return None if \code{overwrite} is enabled, otherwise a \code{numeric}
#'         \code{matrix} containing the intensities after full quantile
#'         normalization.
#' @keywords internal
#'
TileQuantileNorm <- function(int,
                             overwrite){
################################################################################
  stopifnot(!missing(int) && !missing(overwrite))
  
  probes <- nrow(int)
  samples <- ncol(int)
  ord <- matrix(ncol=samples, nrow=probes)
  norm <- rep(0, probes)
  res <- int
  for (i in 1:samples){
    ord[,i] <- order(int[,i])
    norm <- norm + int[ord[,i],i]
  }
  norm <- norm/samples
  for (i in 1:samples){
    res[ord[,i],i] <- norm
  }
  rm(probes, samples)
  rm(ord, norm)
  
  if (overwrite){
    int <<- res
    rm(res)
  }
  else {
    return(res)
  }
}

################################################################################
#'
#' Automatically creates GC content bins.
#'
#' Automatically creates GC content bins under given number of bins. In order to
#' maintain sufficiently large permutation sizes for each GC content bin, the
#' method tries to balance the number of probes within each bin. More precisely,
#' initially a bin is created for each GC content value and iteratively two
#' neighboring bins are merged if their combined number of probes is minimal
#' among all possible neighboring bins. The procedure stops if the given number
#' of \code{gcnum} is reached. In such a way, the most balanced bin sizes can
#' be obtained with the given distribution of GC content values. The method
#' returns 
#'
#' @param  gc A \code{numeric} \code{vector} containing the GC content of each
#'         probe.
#' @param  gcnum The number of preferred GC content bins that should be
#'         automatically created. Note that if \code{gcnum} is higher than the
#'         available number of different GC content values, it will be set to
#'         this maximal number (including a warning message).
#' @return A \code{vector} with the bin boundaries \code{b}. The \code{i}th bin
#'         contains all probes with a GC content within the range
#'         \eqn{[b[i], b[i+1])}.
#' @keywords internal
#'
TileDivideByGC <- function(gc,
                           gcnum=3){
################################################################################
  stopifnot(!missing(gc) || !is.integer(gcnum) || gcnum >= 0)

  ## set lower and upper boundary of
  ## each bin to the available values
  ## of the GC content
  values.high <- as.numeric(names(table(gc)))
  values.low <- values.high
    
  ## record number of occurences
  ## of each available GC content value
  occ <- as.numeric(table(gc))
  if (length(occ) < gcnum){
    TileMsg(file=stderr(), "Warning: The number of selected GC content bins '",
            gcnum, "' is higher than the available GC content values. Setting ",
            "gcnum to '", length(occ), "'.")
    gcnum <- length(occ)
  }

  if (length(occ)-gcnum > 0){
    for (i in 1:(length(occ)-gcnum)){
      ## take two neighboring counts and
      ## min is always index of left border
      min <- which.min(occ[1:(length(occ)-1)] + occ[2:length(occ)])
      ##cat(paste("[", values.low, ",", values.high, "]", sep=""),
      ##"\n", sep="\t")
      ##cat(occ, "\n", sep="\t")
      ##cat("merge: [", values.low[min], ",", values.high[min],"] (",
      ##    occ[min], ") with [", values.low[min+1], ",", values.high[min+1],
      ##    "] (", occ[min+1], ")\n", sep="")
      
      values.high <- values.high[-min]
      values.low <- values.low[-(min+1)]
      occ[min+1] <- occ[min+1] + occ[min]
      occ <- occ[-min]
    }
  }
  ##cat("result:\n")
  ##cat(paste("[", values.low, ",", values.high, "]", sep=""), "\n", sep="\t")
  ##cat(occ, "\n", sep="\t")
  return(c(values.low, Inf))
}

################################################################################
#'
#' Converts start positions to center positions.
#'
#' Converts start positions to center positions with an additional length
#' \code{vector} given. Center positions are rounded off in case of even window
#' length.
#'
#' @param  startpos An \code{integer} \code{vector} containing the start
#'         positions.
#' @param  length An \code{integer} \code{vector} containing the length
#'         information.
#' @return An \code{integer} \code{vector} containing the center position to the
#'         start positions.
#' @keywords internal
#'
TileConvStartToCenter <- function(startpos,
                                  length){
################################################################################
  stopifnot(!missing(startpos) && !missing(length))
  
  off <- floor((length-1)/2)
  return(startpos+off)
}

################################################################################
#'
#' Converts center positions to start positions.
#'
#' Converts center positions to start positions with an additional length
#' \code{vector} given. Center positions were rounded off in case of even window
#' length.
#'
#' @param  centerpos An \code{integer} \code{vector} containing the center
#'         positions.
#' @param  length An \code{integer} \code{vector} containing the length
#'         information.
#' @return An \code{integer} \code{vector} containing the start position to the
#'         center positions.
#' @keywords internal
#'
TileConvCenterToStart <- function(centerpos,
                                  length){
################################################################################
  stopifnot(!missing(centerpos) && !missing(length))
  
  off <- floor((length-1)/2)
  return(centerpos-off)
}


################################################################################
#'
#' Converts center positions to end positions.
#'
#' Converts center positions to end positions with an additional length
#' \code{vector} given. Center positions were rounded off in case of even window
#' length.
#'
#' @param  centerpos An \code{integer} \code{vector} containing one or more
#'         center positions.
#' @param  length An \code{integer} \code{vector} containing one or more length
#'         values. Note that it must contain the same number of values than
#'         \code{centerpos}
#' @return An \code{integer} \code{vector} containing the end position to the
#'         center positions.
#' @keywords internal
#'
TileConvCenterToEnd <- function(centerpos,
                                length){
################################################################################
  stopifnot(!missing(centerpos) && !missing(length) &&
            length(centerpos) == length(length) && length(length) > 0)
  
  return(TileConvCenterToStart(centerpos=centerpos, length=length)+length-1)
}

################################################################################
#'
#' Calculates scores using different functions.
#'
#' Calculates scores using different functions on probe scores.
#'
#' @param  int A \code{numeric} \code{vector} of one or more intensities or fold
#'         changes. It may contain \code{NA} values that are excluded from the
#'         calculation.
#' @param  score.function Function to calculate windows scores over the
#'         \code{log2}-intensities or \code{log2}-fold changes of the
#'         corresponding probes, i.e., arithmetic average
#'         (\code{score.function} = "mean"), arithmetic mean trimmed by the
#'         minimal and maximal value (\code{score.function} = "trimmed"),
#'         or the median (\code{score.function} = "median"). Note that the
#'         definition of trimmed mean differs from the common one with given
#'         percentile ranges. Moreover, the resulting scores with trimmed
#'         mean may only differ from the mean only in case of windows that
#'         contain more than two probes. The latter two scoring functions
#'         are recommended due to their higher robustness against outliers.
#'         However, due to the higher calculation costs, the running time
#'         increases by selecting "trimmed" or "median". Note that the function
#'         is given as \code{character}.
#' @return Returns the calculated value with the selected scoring functions.
#'         Note that windows of length zero are reported with a window score of
#'         \code{NA}. Moreover, probe scores with the value \code{NA} are
#'         excluded from the calculation.
#' @keywords internal
#'
TileCalcScr <- function(int,
                        score.function = c("mean", "trimmed", "median")){
################################################################################
  stopifnot(!missing(int) &&
            !is.na(score.function <- match.arg(score.function)))

  ## remove NA intensities
  int <- int[!is.na(int)]
  
  ## only NA intensities
  if (length(int) == 0){
    return(NA)
  }
  ## mean
  if (score.function == "mean"){
    return(mean(int))
  }
  ## mean trimmed by min and max
  ## (if more than two values)
  if (score.function == "trimmed"){
    res <- sum(int)
    len <- length(int)
    if (length(int) > 2){
      res <- res - max(int) - min(int)
      len <- len - 2
    }
    return(res/len)
  }
  ## median
  if (score.function == "median"){
    return(median(int))
  }
}

################################################################################
#'
#' Wrapper function for calculating window scores using different functions with
#' window indices.
#'
#' Wrapper function for calculating window scores using different functions on
#' probe scores with window indices, a \code{list} storing window start
#' positions. More precisely, \code{win.idx} has the maximal number of probes
#' in a window as length and in its \code{i}th entry, it stores only start
#' positions of windows containing at least \code{(i+1)} probes. In case of
#' short but numerous windows, the window scores can be calculated very
#' efficiently by use of optimized \code{vector} operations. Otherwise, please
#' consider using \code{\link{TileCalcWinScr2}} instead.
#'
#' @param  int A \code{numeric} \code{vector} containing one or more probe
#'         scores (i.e. intensities or log-fold changes).
#' @param  win.idx A \code{list} with the maximal number of probes in a window
#'         as its length. In entry \code{i}, it stores only start positions of
#'         windows containing at least \code{(i+1)} probes.
#' @param  score.function Function to calculate windows scores over the
#'         \code{log2}-intensities or \code{log2}-fold changes of the
#'         corresponding probes, i.e., arithmetic average
#'         (\code{score.function} = "mean"), arithmetic mean trimmed by the
#'         minimal and maximal value (\code{score.function} = "trimmed"),
#'         or the median (\code{score.function} = "median"). Note that the
#'         definition of trimmed mean differs from the common one with given
#'         percentile ranges. Moreover, the resulting scores with trimmed
#'         mean may only differ from the mean only in case of windows that
#'         contain more than two probes. The latter two scoring functions
#'         are recommended due to their higher robustness against outliers.
#'         However, due to the higher calculation costs, the running time
#'         increases by selecting "trimmed" or "median". Note that the function
#'         is given as \code{character}.
#' @return A \code{numeric} \code{vector} containing the calculated window
#'         scores for each possible start position. Note that windows of length
#'         zero are not supported and hence treated as windows of length one.
#'         Moreover, probe scores with the value \code{NA} are excluded from
#'         the calculation.
#' @keywords internal
#'
TileCalcWinScr <- function(int,
                           win.idx,
                           score.function = c("mean", "trimmed", "median")){ 
################################################################################
  stopifnot(!missing(int) && !missing(win.idx) &&
            !missing(score.function) &&
            !is.na(score.function <- match.arg(score.function)))
  
  if (score.function == "mean"){
    return(TileCalcWinMean(int, win.idx))
  }
  if (score.function == "trimmed"){
    return(TileCalcWinTrim(int, win.idx))
  }
  if (score.function == "median"){
    return(TileCalcWinMedian(int, win.idx))
  }
}

################################################################################
#'
#' Wrapper function for calculating window scores using different functions.
#'
#' Wrapper function for calculating window scores using different functions on
#' probe scores with given start positions and window lengths. In case of few
#' but longer windows, this function is superior to \code{\link{TileCalcWinScr}}
#' which is optimized for short windows. 
#'
#' @param  int A \code{numeric} \code{vector} containing one or more probe
#'         scores (i.e. intensities or log-fold changes).
#' @param  win.st An \code{integer} \code{vector} containing one or more window
#'         start positions in \code{int}.
#' @param  win.len An \code{integer} \code{vector} containing one or more window
#'         lengths. It must be of same length as \code{win.st}.
#' @param  score.function Function to calculate windows scores over the
#'         \code{log2}-intensities or \code{log2}-fold changes of the
#'         corresponding probes, i.e., arithmetic average
#'         (\code{score.function} = "mean"), arithmetic mean trimmed by the
#'         minimal and maximal value (\code{score.function} = "trimmed"),
#'         or the median (\code{score.function} = "median"). Note that the
#'         definition of trimmed mean differs from the common one with given
#'         percentile ranges. Moreover, the resulting scores with trimmed
#'         mean may only differ from the mean only in case of windows that
#'         contain more than two probes. The latter two scoring functions
#'         are recommended due to their higher robustness against outliers.
#'         However, due to the higher calculation costs, the running time
#'         increases by selecting "trimmed" or "median". Note that the function
#'         is given as \code{character}.
#' @return A \code{numeric} \code{vector} containing the calculated window
#'         scores for the given start positions. Note that windows of length
#'         zero are reported with a window score of \code{NA}. Moreover, probe
#'         scores with the value \code{NA} are excluded from the calculation.
#' @keywords internal
#'
TileCalcWinScr2 <- function(int,
                            win.st,
                            win.len,
                            score.function = c("mean", "trimmed", "median")){  
################################################################################
  stopifnot(!missing(int) && !missing(win.st) && !missing(win.len) &&
            length(win.st) == length(win.len) && !missing(score.function) &&
            !is.na(score.function <- match.arg(score.function)))
  
  if (score.function == "mean"){
    return(TileCalcWinMean2(int, win.st, win.len))
  }
  if (score.function == "trimmed"){
    return(TileCalcWinTrim2(int, win.st, win.len))
  }
  if (score.function == "median"){
    return(TileCalcWinMedian2(int, win.st, win.len))
  }
}

################################################################################
#'
#' Calculates the average GC content with window indices.
#'
#' Calculates the average window GC content on probe GC contents with window
#' indices, storing window start positions. More precisely, \code{win.idx} has
#' the maximal number of probes in a window as length and in its \code{i}th
#' entry, it stores only start positions of windows containing at least
#' \code{(i+1)} probes. In case of short but numerous windows, the average GC
#' content in windows can be calculated very efficiently by use of optimized
#' \code{vector} operations. Otherwise, please consider using
#' \code{\link{TileCalcWinGC2}} instead.
#'
#' @param  gc A \code{numeric} \code{vector} containing the GC content of one
#'         or more probes. The GC content is the sum of the relative numbers of
#'         Cs and Gs in the probe sequence.
#' @param  win.idx A \code{list} with the maximal number of probes in a window
#'         as its length. In entry \code{i}, it stores only start positions of
#'         windows containing at least \code{(i+1)} probes.
#' @return A \code{numeric} \code{vector} containing the average GC content per
#'         window. Note that windows of length zero are not supported and hence
#'         treated as windows of length one. Moreover, GC content with the value
#'         \code{NA} are excluded from the calculation.
#' @keywords internal
#'
TileCalcWinGC <- function(gc,
                          win.idx){
################################################################################
  stopifnot(!missing(gc) && !missing(win.idx))
  
  res <- rep(0, length(gc))
  res[!is.na(gc)] <- gc[!is.na(gc)]
  len <- rep(0, length(gc))
  len[!is.na(gc)] <- 1

  if (length(win.idx) > 0){
    for (i in 1:length(win.idx)){
      idx <- win.idx[[i]]
      tmp <- gc[idx + i]    
      ##  update gc and length
      len[idx] <- len[idx] + as.integer(!is.na(tmp))
      tmp[is.na(tmp)] <- 0
      res[idx] <- res[idx] + tmp
      ## cleanup
      rm(idx, tmp)
    }
  }
  res[len > 0] <- res[len > 0]/len[len > 0]
  res[len == 0] <- NA
  ## cleanup
  rm(len)
  return(res)
}

################################################################################
#'
#' Calculates the average window GC content.
#'
#' Calculates the average window GC content on probe GC contents with given
#' start positions and window lengths. In case of few but longer windows, this
#' function is superior to \code{\link{TileCalcWinGC}} which is optimized for
#' short windows.
#'
#' @param  gc A \code{numeric} \code{vector} containing the GC content of one
#'         or more probes. The GC content is the sum of the relative numbers of
#'         Cs and Gs in the probe sequence.
#' @param  win.st An \code{integer} \code{vector} containing one or more window
#'         start positions in \code{gc}.
#' @param  win.len An \code{integer} \code{vector} containing one or more window
#'         lengths. It must be of same length as \code{win.st}.
#' @return A \code{numeric} \code{vector} containing the average GC content per
#'         window. Note that windows of length zero are reported with average GC
#'         content of \code{NA}. Moreover, GC content with the value \code{NA}
#'         are excluded from the calculation.
#' @keywords internal
#'
TileCalcWinGC2 <- function(gc,
                           win.st,
                           win.len){
################################################################################
  stopifnot(!missing(gc) && !missing(win.st) && !missing(win.len) &&
            length(win.st) == length(win.len))
  
  res <- rep(0, length(win.st))
  res[win.len > 0 & !is.na(gc[win.st])] <- gc[win.st[win.len > 0 &
                                                     !is.na(gc[win.st])]]
  len <- rep(0, length(win.st))
  len[win.len > 0 & !is.na(gc[win.st])] <- 1
  
  for (i in 1:(max(win.len)-1)){
    idx <- which(win.len > i)
    tmp <- gc[win.st[idx] + i]
    ## update gc and length
    len[idx] <- len[idx] + as.integer(!is.na(tmp))
    tmp[is.na(tmp)] <- 0    
    res[idx] <- res[idx] + tmp
    ## cleanup
    rm(idx, tmp)
  }
  ## calc average GC
  res[len > 0] <- res[len > 0]/len[len > 0]
  res[len == 0] <- NA
  ## cleanup
  rm(len)
  return(res)
}

################################################################################
#'
#' Calculates the minimal q-value within merged windows.
#'
#' Calculates the minimal q-value within merged windows, created by union of
#' overlapping significant windows. Since q-values were estimated only on the
#' original windows, the minimal q-value is assigned to each merged window by
#' considering the q-value of each entirely enclosed window.
#'
#' @param  qval A \code{numeric} \code{vector} containing one or more q-values
#'         that were estimated on the original windows.
#' @param  len An \code{integer} \code{vector} containing the length of one or
#'         more original windows. It must be of same length as \code{qval}.
#' @param  win.st An \code{integer} \code{vector} containing one or more start
#'         positions of the merged (and hence potentially longer) windows.
#' @param  win.len An \code{integer} \code{vector} containing one or more
#'         lengths of the original windows. It must be of same length as
#'         \code{win.st}.
#' @return A \code{numeric} \code{vector} containing the minimal estimated
#'         q-value within each merged window by taking the minimum of all
#'         given q-values of windows that are entirely enclosed in the merged
#'         windows. The result \code{vector} has the same length as
#'         \code{win.st}. Note that windows of length zero are reported with
#'         minimal q-value of \code{NA}. Moreover, q-values with the value
#'         \code{NA} are excluded from the calculation.
#' @keywords internal
#'
TileCalcWinMinQ <- function(qval,
                            len,
                            win.st,
                            win.len){
################################################################################
  stopifnot(!missing(qval) && !missing(len) && !missing(win.st) &&
            !missing(win.len) && length(qval) == length(len) &&
            length(win.st) == length(win.len))
  
  res <- rep(NA, length(win.st))
  res[win.len > 0] <- qval[win.st[win.len > 0]]
  for (i in 1:(max(win.len)-1)){
    ## get idx of all windows where pval
    ## at position i was calculated using
    ## positions which are all enclosed in
    ## these windows
    idx <- which(win.len > i + len[win.st + i] - 1)
    res[idx] <- pmin.int(res[idx], qval[win.st[idx] + i], na.rm=T)
    rm(idx)
  }
  return(res)
}

################################################################################
#'
#' Calculates arithmetic average (mean) on probe scores with window indices.
#'
#' Calculates arithmetic average (mean) on probe scores with window indices, a
#' \code{list} storing window start positions. More precisely, \code{win.idx}
#' has the maximal number of probes in a window as length and in its \code{i}th
#' entry, it stores only start positions of windows containing at least
#' \code{(i+1)} probes. In case of short but numerous windows, the window
#' scores can be calculated very efficiently by use of optimized \code{vector}
#' operations. Otherwise, please consider using \code{\link{TileCalcWinMean2}}
#' instead.
#'
#' @param  int A \code{numeric} \code{vector} containing one or more probe
#'         scores (i.e. intensities or log-fold changes).
#' @param  win.idx A \code{list} with the maximal number of probes in a window
#'         as its length. In entry \code{i}, it stores only start positions of
#'         windows containing at least \code{(i+1)} probes.
#' @return A \code{numeric} \code{vector} containing the arithmetic average of
#'         the window scores for each possible start position. Note that windows
#'         of length zero are not supported and hence treated as windows of
#'         length one. Moreover, probe scores with the value \code{NA} are
#'         excluded from the calculation.
#' @keywords internal
#'
TileCalcWinMean <- function(int, win.idx){
################################################################################
  stopifnot(!missing(int) && !missing(win.idx))
  
  if (all(!is.na(int))){    
    scr <- int
    len <- rep(1, length(int))
    if (length(win.idx) > 0){
      for (i in 1:length(win.idx)){
        len[win.idx[[i]]] <- i + 1
        scr[win.idx[[i]]] <- scr[win.idx[[i]]] + int[win.idx[[i]] + i]
      }
    }
    ## calculate mean
    scr <- scr/len
    ## cleanup
    rm(len)
    stopifnot(all(!is.na(scr)))
    return(scr)
  }
  ## more time and space consuming
  else {
    scr <- rep(0, length(int))
    scr[!is.na(int)] <- int[!is.na(int)]
    len <- rep(0, length(int))
    len[!is.na(int)] <- 1
    if (length(win.idx) > 0){
      for (i in 1:length(win.idx)){
        idx <- win.idx[[i]]
        tmp <- int[idx + i]
        ## update score and length
        len[idx] <- len[idx] + as.integer(!is.na(tmp))
        tmp[is.na(tmp)] <- 0
        scr[idx] <- scr[idx] + tmp
        ## cleanup
        rm(idx, tmp)
      }
    }
    scr[len > 0] <- scr[len > 0]/len[len > 0]
    scr[len == 0] <- NA
    ## cleanup
    rm(len)
    return(scr)
  }
}

################################################################################
#'
#' Calculates arithmetic average (mean) on windows.
#'
#' Calculates arithmetic average (mean) on probe scores with given start
#' positions and window lengths. In case of few but longer windows, this
#' function is superior to \code{\link{TileCalcWinMean}} which is optimized for
#' short windows. 
#'
#' @param  int A \code{numeric} \code{vector} containing one or more probe
#'         scores (i.e. intensities or log-fold changes).
#' @param  win.st An \code{integer} \code{vector} containing one or more window
#'         start positions in \code{int}.
#' @param  win.len An \code{integer} \code{vector} containing one or more window
#'         lengths. It must be of same length as \code{win.st}.
#' @return A \code{numeric} \code{vector} containing the arithmetic average of
#'         the window scores for each possible start position. Note that windows
#'         of length zero are reported with a window score of \code{NA}.
#'         Moreover, probe scores with the value \code{NA} are excluded from
#'         the calculation.
#' @keywords internal
#'
TileCalcWinMean2 <- function(int,
                             win.st,
                             win.len){
################################################################################
  stopifnot(!missing(int) && !missing(win.st) && !missing(win.len) &&
            length(win.st) == length(win.len))
  
  scr <- rep(0, length(win.st))
  scr[win.len > 0 & !is.na(int[win.st])] <- int[win.st[win.len > 0 &
                                                       !is.na(int[win.st])]]
  len <- rep(0, length(win.st))
  len[win.len > 0 & !is.na(int[win.st])] <- 1    
  for (i in 1:(max(win.len)-1)){
    idx <- which(win.len > i)
    tmp <- int[win.st[idx] + i]
    ## update score and length
    len[idx] <- len[idx] + as.integer(!is.na(tmp))
    tmp[is.na(tmp)] <- 0
    scr[idx] <- scr[idx] + tmp
    ## cleanup
    rm(idx, tmp)
  }
  ## calc mean
  scr[len > 0] <- scr[len > 0]/len[len > 0]
  scr[len == 0] <- NA
  ## cleanup
  rm(len)
  return(scr)
}

################################################################################
#'
#' Calculates median on probe scores with window indices.
#'
#' Calculates median on probe scores with window indices, a \code{list} storing
#' window start positions. More precisely, \code{win.idx} has the maximal number
#' of probes in a window as length and in its \code{i}th entry, it stores only
#' start positions of windows containing at least \code{(i+1)} probes. In case
#' of short but numerous windows, the window scores can be calculated very
#' efficiently by use of optimized \code{vector} operations. Otherwise, please
#' consider using \code{\link{TileCalcWinMedian2}} instead.
#'
#' @param  int A \code{numeric} \code{vector} containing one or more probe
#'         scores (i.e. intensities or log-fold changes).
#' @param  win.idx A \code{list} with the maximal number of probes in a window
#'         as its length. In entry \code{i}, it stores only start positions of
#'         windows containing at least \code{(i+1)} probes.
#' @return A \code{numeric} \code{vector} containing the median of the window
#'         scores for each possible start position. Note that windows of length
#'         zero are not supported and hence treated as windows of length one.
#'         Moreover, probe scores with the value \code{NA} are excluded from the
#'         calculation.
#' @import Biobase
#' @keywords internal
#'
TileCalcWinMedian <- function(int,
                              win.idx){
################################################################################
  stopifnot(!missing(int) && !missing(win.idx))
  
  m <- matrix(nrow=length(int), ncol=(length(win.idx) + 1))
  m[,1] <- int
  if (length(win.idx) > 0){
    for (i in 1:length(win.idx)){
      m[win.idx[[i]],(i+1)] <- int[win.idx[[i]] + i]
    }
  }
  med <- rowMedians(m, na.rm=T)
  rm(m)
  return(med)
}

################################################################################
#'
#' Calculates median on windows.
#'
#' Calculates median on probe scores with given start positions and window
#' lengths. In case of few but longer windows, this function is superior to
#' \code{\link{TileCalcWinMedian}} which is optimized for short windows. 
#'
#' @param  int A \code{numeric} \code{vector} containing one or more probe
#'         scores (i.e. intensities or log-fold changes).
#' @param  win.st An \code{integer} \code{vector} containing one or more window
#'         start positions in \code{int}.
#' @param  win.len An \code{integer} \code{vector} containing one or more window
#'         lengths. It must be of same length as \code{win.st}.
#' @return A \code{numeric} \code{vector} containing the median of the window
#'         scores for each possible start position. Note that windows of length
#'         zero are reported with a window score of \code{NA}. Moreover, probe
#'         scores with the value \code{NA} are excluded from the calculation.
#' @keywords internal
#'
TileCalcWinMedian2 <- function(int,
                               win.st,
                               win.len){
################################################################################
  stopifnot(!missing(int) && !missing(win.st) && !missing(win.len) &&
            length(win.st) == length(win.len))
  
  res <- unlist(sapply(1:length(win.st), function(x){
    if (win.len[x] > 0){
      median(int[win.st[x]:(win.st[x]+win.len[x]-1)], na.rm=T)
    }
    else {
      NA
    }}))
  return(res)
}

################################################################################
#'
#' Calculates arithmetic average trimmed by the minimal and maximal value on
#' probe scores with window indices.
#'
#' Calculates arithmetic average trimmed by the minimal and maximal value on
#' probe scores with window indices, a \code{list} storing window start
#' positions. More precisely, \code{win.idx} has the maximal number of probes
#' in a window as length and in its \code{i}th entry, it stores only start
#' positions of windows containing at least \code{(i+1)} probes. In case of
#' short but numerous windows, the window scores can be calculated very
#' efficiently by use of optimized \code{vector} operations. Otherwise, please
#' consider using \code{\link{TileCalcWinTrim2}} instead. Note that the
#' definition of the common'trimmed' mean differs from the common one with given
#' percentile ranges. Moreover, the resulting scores with trimmed mean may only
#' differ from the mean only in case of windows that contain more than two
#' probes.
#'
#' @param  int A \code{numeric} \code{vector} containing one or more probe
#'         scores (i.e. intensities or log-fold changes).
#' @param  win.idx A \code{list} with the maximal number of probes in a window
#'         as its length. In entry \code{i}, it stores only start positions of
#'         windows containing at least \code{(i+1)} probes.
#' @return A \code{numeric} \code{vector} containing the arithmetic mean trimmed
#'         by the minimal and maximal value of the window scores for each
#'         possible start position. Note that windows of length zero are not
#'         supported and hence treated as windows of length one. Moreover, probe
#'         scores with the value \code{NA} are excluded from the calculation.
#' @keywords internal
#'
TileCalcWinTrim <- function(int,
                            win.idx){
################################################################
  stopifnot(!missing(int) && !missing(win.idx))
  
  if (all(!is.na(int))){
    min <- int
    max <- int
    scr <- int
    len <- rep(1, length(int))
    if (length(win.idx) > 0){
      for (i in 1:length(win.idx)){
        idx <- win.idx[[i]]
        tmp <- int[idx + i]
        len[idx] <- i + 1
        ## update min and max
        min[idx] <- pmin.int(min[idx], tmp, na.rm=F)
        max[idx] <- pmax.int(max[idx], tmp, na.rm=F)
        ## sum over all intensities in window
        scr[idx] <- scr[idx] + tmp
        rm(idx, tmp)
      }
    }
    ## trim all windows with at least 3 probes
    ## by min and max
    if (length(win.idx) > 1){
      scr[win.idx[[2]]] <- scr[win.idx[[2]]] - min[win.idx[[2]]] -
        max[win.idx[[2]]]
      len[win.idx[[2]]] <- len[win.idx[[2]]] - 2
    }
    ## calculate mean
    scr <- scr/len
    ## cleanup
    rm(min, max, len)
    stopifnot(all(!is.na(scr)))
    return(scr)
  }
  ## more time and space consuming
  else {
    min <- int
    max <- int
    scr <- rep(0, length(int))
    scr[!is.na(int)] <- int[!is.na(int)]  
    len <- rep(0, length(int))
    len[!is.na(int)] <- 1
    if (length(win.idx) > 0){
      for (i in 1:length(win.idx)){
        idx <- win.idx[[i]]
        tmp <- int[idx + i]
        ## update length (non-NA probes per window)
        len[idx] <- len[idx] + as.integer(!is.na(tmp))
        ## update min and max
        min[idx] <- pmin.int(min[idx], tmp, na.rm=T)
        max[idx] <- pmax.int(max[idx], tmp, na.rm=T)
        ## sum over all intensities in window
        tmp[is.na(tmp)] <- 0
        scr[idx] <- scr[idx] + tmp
        ## cleanup
        rm(idx, tmp)
      }
    }
    ## trim all windows with at least 3 probes
    ## by min and max
    idx <- which(len > 2)
    scr[idx] <- scr[idx] - min[idx] - max[idx]
    len[idx] <- len[idx] - 2
    rm(idx)
    ## calculate mean
    scr[len > 0] <- scr[len > 0]/len[len > 0]
    scr[len == 0] <- NA
    ## cleanup
    rm(min, max, len)
    return(scr)
  }
}

################################################################################
#'
#' Calculates arithmetic average trimmed by the minimal and maximal value on
#' probe scores on windows.
#'
#' Calculates arithmetic average trimmed by the minimal and maximal value on
#' probe scores on probe scores with given start positions and window lengths.
#' In case of few but longer windows, this function is superior to
#' \code{\link{TileCalcWinTrim}} which is optimized for short windows. 
#'
#' @param  int A \code{numeric} \code{vector} containing one or more probe
#'         scores (i.e. intensities or log-fold changes).
#' @param  win.st An \code{integer} \code{vector} containing one or more window
#'         start positions in \code{int}.
#' @param  win.len An \code{integer} \code{vector} containing one or more window
#'         lengths. It must be of same length as \code{win.st}.
#' @return A \code{numeric} \code{vector} containing the arithmetic mean trimmed
#'         by the minimal and maximal value of the window scores for each
#'         possible start position. Note that windows of length zero are
#'         reported with a window score of \code{NA}. Moreover, probe scores
#'         with the value \code{NA} are excluded from the calculation.
#' @keywords internal
#'
TileCalcWinTrim2 <- function(int,
                             win.st,
                             win.len){
################################################################################
  stopifnot(!missing(int) && !missing(win.st) && !missing(win.len) &&
            length(win.st) == length(win.len))
  
  min <- int[win.st]
  max <- int[win.st]
  scr <- rep(0, length(win.st))
  scr[win.len > 0 & !is.na(int[win.st])] <- int[win.st[win.len > 0 &
                                                       !is.na(int[win.st])]]
  len <- rep(0, length(win.st))
  len[win.len > 0 & !is.na(int[win.st])] <- 1    
  for (i in 1:(max(win.len)-1)){
    idx <- which(win.len > i)
    tmp <- int[win.st[idx] + i]
    ## update length (non-NA probes per window)
    len[idx] <- len[idx] + as.integer(!is.na(tmp))
    ## update min and max
    min[idx] <- pmin.int(min[idx], tmp, na.rm=T)
    max[idx] <- pmax.int(max[idx], tmp, na.rm=T)
    ## sum over all intensities in window
    tmp[is.na(tmp)] <- 0
    scr[idx] <- scr[idx] + tmp
    ## cleanup
    rm(idx, tmp)
  }
  ## trim all windows with at least 3 probes
  ## by min and max
  idx <- which(len > 2)
  scr[idx] <- scr[idx] - min[idx] - max[idx]
  len[idx] <- len[idx] - 2
  ## calculate mean
  scr[len > 0] <- scr[len > 0]/len[len > 0]
  scr[len == 0] <- NA
  ## cleanup
  rm(min, max, len, idx)
  return(scr)
}

################################################################################
#'
#' Converts relative to absolute paths.
#'
#' Converts a given \code{character} \code{vector} of relative path to absolute
#' paths in their canonical form by use of the current working directory and
#' some path normalizations, i.e., no ./, ../ nor symbolic links under unix.
#'
#' @param  filename A \code{character} \code{vector} of one or more relative
#'         filenames. Note that absolute paths will not be changed.
#' @return A \code{character} \code{vector} of same length as \code{filename}
#'         of absolute paths in their canonical form.
#' @import utils
#' @keywords internal
#'
TileConvPath <- function(filename){
################################################################################
  stopifnot(!missing(filename))
  if (is.null(filename)){
    return("NULL")
  }
  
  ## normalize path (no ./, ../ nor symbolic links)
  res <- normalizePath(filename)
  
  ##for (i in 1:length(res)){
    ## relative path
  ##  if (!grepl("^/", res[i])){
  ##    res[i] <- file.path(getwd(), res[i])
  ##  }
  ##}

  ##res <- normalizePath(res)
  
  return(res)
}


################################################################################
#'
#' Prints message with standardized prefix.
#'
#' Prints message with standardized prefix consisting of package name and
#' the current timestamp.
#'
#' @param  file Filename or stream to which the output will be printed.
#' @param  ... One or more \code{R} objects, to be converted to \code{character}
#'         \code{vectors}.
#' @return None.
#' @keywords internal
#'
TileMsg <- function(..., file=stdout()){
################################################################################
  cat(file=file, "[", getPackageName(), "] ",
      date(), ": ", ..., "\n", sep="")
}
