
################################################################################
#'
#' Checks parameter for exported function \code{\link{TileAnalysis}}.
#'
#' Checks parameter for exported function \code{\link{TileAnalysis}} whether
#' they are missing, have a incorrect data type, or are obviously not correct,
#' e.g., filename parameter refers to non-existing file.
#'
#' @param  data A \code{data.frame} with non-zero rows and the columns fullname
#'         (\code{factor} or \code{character}), name (\code{factor} or
#'         \code{character}), centerpos (\code{integer}), probelength
#'         (\code{integer}), intensity (\code{numeric}), and gc
#'         (\code{numeric}) with only non-NA values. Moreover, the gc must
#'         have values in the range [0,1].
#' @param  noofperms An \code{integer} value greater than zero.
#' @param  winsize An \code{integer} value greater than zero.
#' @param  qvalue A \code{numeric} value between zero and one (both boundaries
#'         included).
#' @param  gcmode A \code{character} value from the set of valid values
#'         ("fixed", "automatic").
#' @param  gcnum An \code{integer} value. If \code{gcmode} is set to "fixed"
#'         only values from the set (1, 2, 3, 4, or 5) are valid.
#' @param  score.function A \code{character} value from the set of valid values
#'         ("mean", "trimmed", or "median").
#' @param  randomize A \code{logical} value.
#' @param  diff A \code{logical} value.
#' @param  diff.variant A \code{character} value from the set of valid values
#'         ("A" or "B"). Only required if \code{diff} is enabled.
#' @param  regions.filename Filename as a \code{character} to existing file or
#'         set to \code{NULL}.
#' @param  zscore A \code{logical} value.
#' @param  verbose A \code{logical} value.
#' @return None but the given paramters into their correct data type if
#'         necessary and possible.
#' @keywords internal
#'
TileAnalysisCheck <- function(data,
                              noofperms,
                              winsize,
                              qvalue,
                              gcmode,
                              gcnum,
                              score.function,
                              randomize,
                              diff,
                              diff.variant,
                              regions.filename,
                              zscore,
                              verbose){
################################################################################
  if (!missing(data)){
    if (!is.data.frame(data) || nrow(data) == 0 ||
        any(!c("fullname", "name", "centerpos", "probelength",
               "intensity", "gc") %in% colnames(data)) ||
        !(is.factor(data$fullname) || is.character(data$fullname)) ||
        !(is.factor(data$name) || is.character(data$name)) ||
        !is.integer(data$centerpos) || !is.integer(data$probelength) ||
        !is.numeric(data$intensity) || !is.numeric(data$intensity) ||
        any(is.na(data)) || any(data$gc < 0) || any(data$gc > 1)){
      stop("The input data is invalid.")
    }
  }

  if (missing(noofperms) || !(is.integer(noofperms) || is.numeric(noofperms)) ||
      round(noofperms) != noofperms || length(noofperms) != 1 ||
      noofperms <= 0){
    stop("The parameter 'noofperms' is missing or invalid.")
  }
  if (missing(winsize) || !(is.integer(winsize) || is.numeric(winsize))||
      round(winsize) != winsize || length(winsize) != 1 || winsize <= 0){
    stop("The parameter 'winsize' is missing or invalid.")
  }
  if (missing(qvalue) || !(is.integer(qvalue) || is.numeric(qvalue)) ||
      length(qvalue) != 1 || qvalue < 0 || qvalue > 1){
    stop("The parameter 'qvalue' is missing or invalid.")
  }
  if (missing(gcmode) || !is.character(gcmode) || length(gcmode) != 1){
    stop("The parameter 'gcmode' is missing or invalid.")
  }
  else {
    if (! (gcmode %in% c("fixed", "automatic"))){
      stop("The parameter 'gcmode' can only be set to \"fixed\" or ",
           "\"automatic\".")
    }
  }
  if (missing(gcnum) || !(is.integer(gcnum) || is.numeric(gcnum)) ||
      round(gcnum) != gcnum || length(gcnum) != 1){
    stop("The parameter 'gcnum' is missing or invalid.")
  }
  else {
    if (gcmode == "fixed" && !(gcnum %in% c(1,2,3,4,5))){
      stop("The parameter 'gcnum' can only be set to 1, 2, 3, or 4 ",
           "if 'gcmode' is set to \"fixed\".")
    }
  }
  if (missing(score.function) || !is.character(score.function) ||
      length(score.function) != 1){
    stop("The parameter 'score.function' is missing or invalid.")
  }
  else {
    if (!(score.function %in% c("mean", "trimmed", "median"))){
      stop("The parameter 'score.function' can only be set to \"mean\", ",
           "\"trimmed\", or \"median\".")
    }
  }
  if (missing(randomize) || !is.logical(randomize) || length(randomize) != 1){
    stop("The parameter 'randomize' is missing or invalid.")
  }
  if (missing(diff) || !is.logical(diff) || length(diff) != 1){
    stop("The parameter 'diff' is missing or invalid.")
  }
  if (diff){
    if (missing(diff.variant) || !is.character(diff.variant) ||
        length(diff.variant) != 1){
      stop("The parameter 'diff.variant' is missing or invalid.")
    }
    if (!(diff.variant %in% c("A", "B"))){
      stop("The parameter 'diff.variant can only be set to \"A\" or \"B\".")
    }
  }
  if (missing(regions.filename) ||
      (!is.null(regions.filename) && !is.character(regions.filename)) ||
      (!is.null(regions.filename) && length(regions.filename) != 1)){
    stop("The parameter 'regions.filename' is missing or invalid.")
  }
  else {
    if (!is.null(regions.filename) && !file.exists(regions.filename)){
      stop("BED-formatted regions file not found.")
    }
  }
  if (missing(zscore) || !is.logical(zscore) || length(zscore) != 1){
    stop("The parameter 'zscore' is missing or invalid.")
  }
  if (missing(verbose) || !is.logical(verbose) || length(zscore) != 1){
    stop("The parameter 'verbose' is missing or invalid.")
  }
}

################################################################################
#'
#' The statistical analysis of tiling array expression data.
#'
#' Executes the statistical analysis of tiling array expression data that
#' identifies (differential) expression as significant changes from the
#' background distribution while considering sequence-specific affinities as
#' well as cross-hybridization. This method returns a \code{list} with two
#' \code{data.frames}: one containing the information on the analyzed window
#' including their estimated z-score if \code{zscore} is enabled while the other
#' one comprise the significantly (differentially) expressed segments.
#'
#' @param  data A \code{data.frame} containing information on all probes,
#'         i.e., the full name of the reference sequence (organism abbreviation
#'         and chromosome name), the chromosome name, the probe center position,
#'         the length of the probe sequence, the GC content of the probe
#'         sequence, and the probe score. This \code{data.frame} is reported by
#'         the functions \code{\link{TileReadCel}}, \code{\link{TileReadSignal}},
#'         or \code{\link{TileReadCustom}}.
#' @param  noofperms Number of permutations used to sample the background
#'         distribution. With higher number of permutations, the statistical
#'         significance of windows can be assessed more precisely in particular
#'         with more restrictive significance thresholds, i.e., low values of
#'         the \code{qvalue} parameter. Moreover, in case of the differential
#'         analysis (\code{diff} is enabled), the number of permutations should
#'         be increased to sample the background distribution more accurately.
#'         In general, \code{noofperm} is recommended to be set to \code{10000}
#'         and \code{100000} in case of expression and differential expression
#'         analyses, respectively.
#' @param  winsize Maximal width of the windows that are being statistically
#'         assessed. The width is defined as the difference in the genomic
#'         center positions of the first and last enclosed probe. Due to
#'         gaps in the commonly uniform distribution of probes over the
#'         genomic sequence. The analyzed windows may be considerably shorter
#'         than the defined \code{winsize} or may even consist of only one
#'         probe.
#' @param  qvalue Maximal permitted q-value that is applied in the statistical
#'         analysis. Hence, windows with a q-value above the given value will
#'         not be included in the returned \code{data.frame}.
#' @param  gcmode Mode of GC content binning. In case of \code{gcmode} set to
#'         "fixed", the classification of probes in bins was predefined
#'         considering the GC content effect on probe intensities on the
#'         Affymetrix tiling array 1.0R platform. In this case, only the
#'         following values are permitted: 1, 2, 3, 4, or 5. By setting
#'         \code{gcmode} to "automatic", the binning is done automatically
#'         solely on the distribution of the GC content of the probes in order
#'         to obtain GC content bins that are optimally balanced in terms of
#'         their sizes.
#' @param  gcnum Number of different GC content bins where probes within each
#'         bin have a similar expected sequence-specific affinity and are
#'         permuted independently from each other. Accordingly, intensities
#'         of probes that belong to different affinity bins must not be
#'         interchanged. Due to the trade-off between the reduction of the
#'         sequence-specific effect and the maintenance of sufficiently large
#'         permutation bins, three GC content bins are recommended in the
#'         expression analysis. In case of \code{gcmode} set to "fixed", the
#'         classification of probes in bins was predefined considering the
#'         GC content effect on probe intensities on the Affymetrix tiling
#'         array 1.0R platform. In this case, only the following values are
#'         permitted: 1, 2, 3, 4, or 5. By setting \code{gcmode} to "automatic",
#'         the binning is done automatically solely on the distribution of the
#'         GC content of the probes in order to obtain GC content bins that are
#'         optimally balanced in terms of their sizes. Note that the
#'         \code{gcnum} is set to one in case of differential expression
#'         analysis (\code{diff} enabled) since sequence-specific effect cancel
#'         out and affinity binning is rendered unnecessary.
#' @param  score.function Function to calculate windows scores over the
#'         scores of the corresponding probes, i.e., arithmetic average
#'         (\code{score.function} = "mean"), arithmetic mean trimmed by the
#'         minimal and maximal value (\code{score.function} = "trimmed"),
#'         or the median (\code{score.function} = "median"). Note that the
#'         definition of trimmed mean differs from the common one with given
#'         percentile ranges. Moreover, the resulting scores with trimmed
#'         mean may differ from the mean only in case of windows that contain
#'         more than two probes. The latter two scoring functions are
#'         recommended due to their higher robustness against outliers. However,
#'         due to the higher calculation costs, the running time increases by
#'         selecting "trimmed" or "median". Note that the function is given as
#'         \code{character}.
#' @param  randomize Indicates whether an additional permutation is applied
#'         prior to the calculation of original window scores. It is a
#'         possiblity to roughly estimate the false positive rate since under
#'         the assumption of mostly unexpressed probes no window over permuted
#'         intensities is expected to differ significantly from the background
#'         distribution.
#' @param  diff Indicates whether differential expression analysis is applied.
#' @param  diff.variant The variants of the differential expression analysis
#'         differ in score calculation, in the permutation procedure as well as
#'         in their assignment of statistical significance to windows. The
#'         \code{diff.variant} A is similar to the normal expression analysis
#'         but two-tailed p-values are estimated to regard both regulation
#'         directions, up and down. The multiple testing correction is then
#'         adjusted to account for these additional comparisons. The
#'         \code{diff.variant} B assumes that entire windows are either up-
#'         or down-regulated between conditions. The presumed direction of
#'         regulation is initially assigned to each window on the basis of its
#'         score. Subsequently, all converse probes, i.e., probes with negative
#'         score within positive windows or vice versa, are ignored and neigther
#'         permuted nor incorporated into the score calculation. Consequently,
#'         positive and negative windows are compared to different background
#'         distributions. The p-value estimation and correction is done
#'         equivalent as in the case of the normal expression analyses. Both
#'         variants produce fairly similar results while the variant B is
#'         slightly superior in its performance and hence recommended.
#' @param  regions.filename Filename of BED-formatted file that contains regions
#'         to which the (differential) expression analysis should be limited to.
#'         Hence, only windows entirely enclosed in the union of these regions
#'         are statistically evaluated. Commonly, this parameter is used in
#'         order to identify highly and differentially expression (highdiff)
#'         regions by restricting the differential expression analysis
#'         (\code{diff} enabled) to regions identified as highly expressed in
#'         either one of the corresponding cellular conditions.
#' @param  zscore Indicates whether z-scores, i.e., normalized window scores,
#'         should be calculated by use of the sampled background. More
#'         precisely, the window z-score \code{z} is calculated by
#'         \eqn{z = \frac{x - \mu}{\sigma}} where \code{x} is the window score
#'         and \eqn{\mu} and \eqn{\sigma} are the mean and standard deviation
#'         of the permuted window scores, respectively. Note that negative
#'         z-scores indicate down-regulation and positive indicate up-regulated
#'         regions since z-score are bounded to zero from below or from above in
#'         case of up- or down-regulation, respectively. Hence, using common
#'         expression analysis (if \code{diff} is \code{FALSE}), the z-score
#'         cannot be negative.
#' @param  verbose Indicates whether information on progress are printed.
#' @return Returns a \code{list} with two \code{data.frames}: one containing
#'         the analyzed windows including the calculated z-score and the other
#'         one comprise the significantly expressed segments. The first
#'         \code{list} entry is \code{NULL} if \code{zscore} is \code{FALSE}.
#'         Otherwise, the entry keeps the z-score \code{data.frame} containing
#'         name of reference sequence, start and end position, description,
#'         estimated z-score, and `+' as strand for each analyzed window. The
#'         description is an underscore-delimitted string of the number of
#'         covered probes, the average GC content of their sequences, the window
#'         q-value multiplied by 100, and the window score calculated with the
#'         given \code{scoring.function} on the probe scores. The second
#'         \code{list} entry is a \code{data.frame} that comprises the
#'         significantly (differentially) expressed segments including name of
#'         reference sequence, start and end position, description, score that
#'         is uniformly set to `10' and `+' as strand. The description is an
#'         underscore-delimitted string of the name of reference sequence,
#'         the start and end position, the number of probes covered by the segment,
#'         the average GC content of their sequences, the minimal q-value of
#'         windows that were merged into the segment (multiplied by 100), and
#'         the segment score calculated with the given \code{scoring.function}
#'         on the covered probes.
#' @import affxparser
#' @import intervals
#' @import genomeIntervals
#' @import multtest
#' @include TileHelper.R
#' @keywords internal
#'
TileAnalysis <- function(data,
                         noofperms,
                         winsize,
                         qvalue,
                         gcmode = "fixed",
                         gcnum,
                         score.function = "trimmed",
                         randomize = FALSE,
                         diff,
                         diff.variant = "B",
                         regions.filename = NULL,
                         zscore = TRUE,
                         verbose = FALSE){
################################################################################
  res <- NULL

  ## check given parameters
  TileAnalysisCheck(data=data, noofperms=noofperms, winsize=winsize,
                    qvalue=qvalue, gcmode=gcmode, gcnum=gcnum,
                    score.function=score.function, randomize=randomize,
                    diff=diff, diff.variant=diff.variant,
                    regions.filename=regions.filename, zscore=zscore,
                    verbose=verbose)
  ## additional check
  if (diff && gcnum > 1){
    TileMsg(file=stderr(), "Warning: GC content binning disregarded in case ",
            "of the differential expression analysis.")
    gcnum <- 1
  }

  if (verbose){
    TileMsg("Statistical analysis started with noofperms=", noofperms,
            ", winsize=", winsize, ", qvalue=", qvalue, ", gcmode=", gcmode,
            ", gcnum=", gcnum, ", score.function=", score.function,
            ", randomize=", randomize, ", diff=", diff,
            ifelse(diff, paste(", diff.variant=", diff.variant, sep=""), ""),
            ", regions.filename=", TileConvPath(filename=regions.filename),
            ", zscore=", zscore)
  }

  ## sort input data by chromosome, probe centerpos, and probe length
  data <- data[order(data$name, data$centerpos, data$probelength),]

  if (diff){
    ## split according to plus and minus probes
    ## NOTE: ivals.idx only necessary if performance
    ## improvement is disabled
    if (diff.variant != "A"){
      ivals.idx <- c(list(which(data$intensity < 0)),
                     list(which(data$intensity >= 0)))
    }
    else {
      ivals.idx <- list(1:length(data$intensity))
    }
  }
  ######################### GC CONTENT BINNING #################################
  ## define GC content bins for shuffling
  ## bin[i] = [ival[i], ival[i+1])
  else {
    ## fixed GC content binning that was manually determined
    ## considering the GC content effect of probe intensities
    ## on the Affy 1.0R tiling array platform
    if (gcmode == "fixed"){
      if (gcnum == 0 || gcnum == 1){
        ivals <- c(0,Inf) ## no classes
      }
      else if (gcnum == 2){
        ivals <- c(0, 0.52, Inf) ## 2 classes
      }
      else if (gcnum == 3){
        ivals <- c(0, 0.36, 0.52, Inf) ## 3 classes
      }
      else if (gcnum == 4){
        ivals <- c(0, 0.36, 0.52, 0.68, Inf) ## 4 classes
      }
      else if (gcnum == 5){
        ivals <- c(0, 0.32, 0.40, 0.44, 0.52, Inf) ## 5 classes
      }
    }
    ## automatic selection of
    ## GC content bins
    else {
      ivals <- TileDivideByGC(gc=data$gc, gcnum=gcnum)
    }
    if (verbose){
      TileMsg("GC content bins with gcmode=", gcmode, ": ",
              paste("[", round(ivals[1:(length(ivals)-1)], digits=3), ",",
                    round(ivals[2:length(ivals)], digits=3), ")",
                    sep="", collapse=" "), ".")
    }
    ivals.idx <- NULL
    un <- NULL
    for (i in 1:(length(ivals) - 1)){
      ## select idx corresponding to current bin
      ivals.idx <- c(ivals.idx, list(which(data$gc >= ivals[i] &
                                           data$gc < ivals[i+1])))
      ## no position to be shuffled
      if (length(ivals.idx[[i]]) == 0){
        TileMsg(file=stderr(), "Warning: No probe within GC content bin [",
                round(ivals[i], digits=3), ",",round(ivals[i+1], digits=3), ")")
        next
      }
      ## already shuffled position?
      stopifnot(length(intersect(un, ivals.idx[[i]])) == 0)
      un <- union(un, ivals.idx[[i]])
    }
    stopifnot(length(un) == length(data$gc))
    rm(un)
  }

  ######################## FPR ESTIMATION ######################################
  ## shuffling before evaluation procedure as control
  if (randomize){
    int.sh <- rep(NA, length(data$intensity))
    for (j in 1:length(ivals.idx)){
      int.sh[ivals.idx[[j]]] <- sample(data$intensity[ivals.idx[[j]]],
                                       length(ivals.idx[[j]]), replace=FALSE)
    }
    if (any(is.na(int.sh))){
      stop("Not every position has been sampled.")
    }
    data$intensity <- int.sh
    rm(int.sh)
  }

  ######################## DEFINE WINDOWS ######################################
  ## find maximal probe count within window size
  if (verbose){
    TileMsg("Windowing tiling array")
  }
  max <- 1
  while(TRUE){
    win.max <- data$centerpos[(max + 1):length(data$centerpos)] -
      data$centerpos[1:(length(data$centerpos) - max)]
    if (length(which(win.max <= winsize &
          data$fullname[(max+1):length(data$fullname)] ==
          data$fullname[1:(length(data$fullname) - max)])) == 0){
      break
    }
    max <- max + 1
    rm(win.max)
  }
  ## win.idx[[i]] = indices of probes having at least
  ## (i + 1) probes in the window
  win.idx <- NULL
  ## find maximal probe count for each probe
  ## (minimum is always 1)
  win.len <- rep.int(1, length(data$centerpos))
  if (max > 1){
    un <- NULL
    for (i in max:2){
      win <- data$centerpos[i:length(data$centerpos)] -
        data$centerpos[1:(length(data$centerpos) - i + 1)]
      tmp <- which(win <= winsize & data$fullname[i:length(data$fullname)] ==
                   data$fullname[1:(length(data$fullname) - i + 1)])
      win.len[setdiff(tmp, un)] <- i
      un <- union(un, tmp)
      win.idx <- c(list(un), win.idx)
      rm(win)
    }
    rm(un)
  }

  ######################### MASK REGIONS #######################################
  if (!is.null(regions.filename)){
    if (verbose){
      TileMsg("Reducing windows to given regions")
    }
    regions <- read.table(regions.filename, stringsAsFactors=F)
    regions.ival <- new("Genome_intervals", as.matrix(regions[,2:3]),
                        closed=TRUE, annotation=data.frame(seq_name=regions[,1],
                                       inter_base=FALSE))
    rm(regions)
    win.ival <- new("Genome_intervals", matrix(ncol=2, byrow=FALSE,
       c(TileConvCenterToStart(centerpos=data$centerpos,
                               length=data$probelength),
         TileConvCenterToEnd(centerpos=data$centerpos[1:nrow(data)+win.len-1],
                             length=data$probelength[1:nrow(data)+win.len-1]))),
       closed=TRUE, annotation=data.frame(seq_name=data$name, inter_base=FALSE))

    o <- interval_overlap(win.ival, interval_complement(regions.ival))
    idx <- which(unlist(lapply(o, length)) > 0)
    ## window idx is excluded and length is set to zero
    ## if window overlaps with any given region
    win.len[idx] <- 0
    ## abort if nothing left to calculate
    if (all(win.len == 0)){
      cat(file=stderr(), "No windows left after reducing. Exit forced.")
      stop()
    }
    if (length(win.idx) > 0){
      for (i in 1:length(win.idx)){
        win.idx[[i]] <- setdiff(win.idx[[i]], idx)
      }
    }
    rm(regions.ival, win.ival, o, idx)
  }

  ################### CALCULATE ORIGINAL WINDOW SCORE ##########################
  if (verbose){
    TileMsg("Calculating original window scores")
  }
  ## score calculation according to defined function
  ## consider only windows with non-zero length
  win.scr <- TileCalcWinScr(int=data$intensity, win.idx=win.idx,
                            score.function=score.function)
  stopifnot(all(!is.na(win.scr)))

  if (!diff || diff.variant == "A"){
    win.idx2 <- NULL
    for (i in 1:max(win.len)){
      win.idx2 <- c(win.idx2, list(which(win.len == i)))
    }
  }
  ## different handling of two-sided distributions
  ## with diff.variant != "A"
  else {
    ## split windows into plus and minus ones (according to score)
    win.plus <- which(win.scr >= 0)
    win.minus <- which(win.scr < 0)

    ## split win.idx into plus and minus
    win.idx.plus <- NULL
    win.idx.minus <- NULL
    if (length(win.idx) > 0){
      for (i in 1:length(win.idx)){
        win.idx.plus <- c(win.idx.plus,
                          list(intersect(win.plus, win.idx[[i]])))
        win.idx.minus <- c(win.idx.minus,
                           list(intersect(win.minus, win.idx[[i]])))
      }
    }

    ## consider plus or minus probes only
    int.plus <- which(data$intensity >= 0)
    int.minus <- which(data$intensity < 0)

    ## rescore windows taking only plus or minus intensities
    if (length(int.plus) > 0){
      tmp <- data$intensity
      tmp[int.minus] <- NA
      win.scr[win.plus] <- TileCalcWinScr(int=tmp, win.idx=win.idx.plus,
                                        score.function=score.function)[win.plus]
    }
    if (length(int.minus) > 0){
      tmp <- data$intensity
      tmp[int.plus] <- NA
      win.scr[win.minus] <- TileCalcWinScr(int=tmp, win.idx=win.idx.minus,
                                       score.function=score.function)[win.minus]
    }
    stopifnot(all(!is.na(win.scr)))

    ## keep reduced number of probes for each window
    ## after excluding not fitting probes
    ## --> ignore such windows in evaluation
    win.len2 <- rep(0, length(win.len))
    for (i in 1:max(win.len)){
      idx <- intersect(win.plus, which(win.len >= i))
      win.len2[idx] <- win.len2[idx] +
        ifelse(data$intensity[idx+i-1] >= 0, 1, 0)
      idx <- intersect(win.minus, which(win.len >= i))
      win.len2[idx] <- win.len2[idx] +
        ifelse(data$intensity[idx+i-1] < 0, 1, 0)
      rm(idx)
    }
    stopifnot(all(win.len2[win.len > 0] > 0) &&
              all(win.len2[win.len == 0] == 0) &&
              all(win.len >= win.len2))

    win.idx2 <- NULL
    for (i in 1:max(win.len2)){
      idx <- which(win.len2 == i)
      win.idx2 <- c(win.idx2, list(list(plus=intersect(win.plus, idx),
                                        minus=intersect(win.minus, idx))))
      rm(idx)
    }
    rm(tmp)
  }
  ## find number of sampled windows whose intensity sum
  ## is greater than the original intensity sum
  win.cnt <- rep.int(0, length(win.scr))
  ## second counter for differential expression
  ## with diff.variant == 1
  if (diff && diff.variant == "A"){
    win.cnt2 <- rep.int(0, length(win.scr))
  }

  if (zscore){
    win.scr.sqmu <- rep.int(0, length(win.scr))
    win.scr.mu <- rep.int(0, length(win.scr))
  }

  ############################ SHUFFLING #######################################
  if (verbose){
    TileMsg("Shuffling started")
  }
  for (i in 1:noofperms){
    if (!diff && length(ivals.idx) > 1){
      int.sh <- rep(NA, length(data$intensity))
      ## shuffle intensities respectively to the classification
      for (j in 1:length(ivals.idx)){
        if (length(ivals.idx[[j]]) > 0){
          int.sh[ivals.idx[[j]]] <- sample(x=data$intensity[ivals.idx[[j]]],
                                           size=length(ivals.idx[[j]]),
                                           replace=FALSE)
        }
      }
      if (any(is.na(int.sh))){
        stop("Not every position has been sampled.")
      }
      ## calculate score for each window with non-zero length
      win.scr.sh <- TileCalcWinScr(int=int.sh, win.idx=win.idx,
                                   score.function=score.function)
      stopifnot(all(!is.na(win.scr.sh)))
    }
    ## performance speed up if no
    ## GC classification has been selected
    ## --> sampling up to maximal window length sufficient
    else {
      win.scr.sh <- rep(0, length(win.scr))

      if (!diff || diff.variant == "A"){
        int.sh <- sample(x=data$intensity, size=length(win.idx2), replace=FALSE)
        for (j in 1:length(win.idx2)){
          ## win.idx2[[j]] stores indices with win.len == j
          if (length(win.idx2[[j]]) > 0){
            win.scr.sh[win.idx2[[j]]] <- TileCalcScr(int=int.sh[1:j],
                                                  score.function=score.function)
          }
        }
      }
      else {
        int.sh <- NULL; int.sh.minus <- NULL
        ## sample with positive and negative background seperatively
        if (length(int.plus) > 0){
          int.sh <- sample(x=data$intensity[int.plus],
                           size=length(win.idx2), replace=FALSE)
        }
        if (length(int.minus) > 0){
          int.sh.minus <- sample(x=data$intensity[int.minus],
                                 size=length(win.idx2), replace=FALSE)
        }
        for (j in 1:length(win.idx2)){
          ## win.idx2[[j]] stores in two lists (plus and minus) indices
          ## of positive or negative windows with reduced length win.len2 == j
          if (length(win.idx2[[j]]$plus) > 0){
            win.scr.sh[win.idx2[[j]]$plus] <- TileCalcScr(int=int.sh[1:j],
                                              score.function=score.function)
          }
          if (length(win.idx2[[j]]$minus) > 0){
            win.scr.sh[win.idx2[[j]]$minus] <- TileCalcScr(int=int.sh.minus[1:j],
                                               score.function=score.function)
          }
        }
        rm(int.sh.minus)
      }
    }
    ## comparison with background scores
    if (!diff){
      win.cnt <- win.cnt + as.integer(win.scr.sh > win.scr)
    }
    else {
      if (diff.variant == "A"){
        win.cnt <- win.cnt + as.integer(win.scr.sh > win.scr)
        win.cnt2 <- win.cnt2 + as.integer(win.scr.sh < win.scr)
      }
      else {
        ## only compare with respective background
        win.cnt[win.plus] <- win.cnt[win.plus] +
          as.integer(win.scr.sh[win.plus] > win.scr[win.plus])
        win.cnt[win.minus] <- win.cnt[win.minus] +
          as.integer(win.scr.sh[win.minus] < win.scr[win.minus])
      }
    }
    ## calculate mean and standard deviation of score
    ## for z-score
    if (zscore){
      win.scr.sqmu <- win.scr.sqmu + (win.scr.sh)^2
      win.scr.mu <- win.scr.mu + win.scr.sh
    }
    rm(int.sh, win.scr.sh)

    ## report progress
    if (verbose){
      if (i %in% seq(0, noofperms, ceiling(noofperms/10))){
        perc <- 100*(i/noofperms)
        TileMsg(i, " of ", noofperms, " (", round(perc, digits=1),
                "%) shufflings finished")
      }
    }
  } ## end of for (i in 1:noofperms)
  rm(ivals.idx)


  ########################## Q-VALUE CALCULATION ###############################
  if (verbose){
    TileMsg("Estimating q-values")
  }
  if (!diff){
    win.pval <- win.cnt/noofperms
  }
  else {
    if (diff.variant == "A"){
      win.pval <- 2 * pmin(win.cnt, win.cnt2)/noofperms
    }
    else {
      win.pval <- win.cnt/noofperms
    }
  }
  ## correcting pvalues with respect to multiple testing
  ## requires library(multtest), only using non-zero
  ## length windows
  list <- mt.rawp2adjp(rawp=win.pval[win.len > 0], proc="BH")
  win.qval <- rep(NA, length(win.scr))
  win.qval[win.len > 0] <- list$adjp[order(list$index),2]
  rm(list)
  win.res <- which(!is.na(win.qval) & win.qval < qvalue)

  ######################### Z-VALUE CALCULATION ################################
  if (zscore){
    if (verbose){
      TileMsg("Estimating z-scores")
    }
    ## calculate mean of shuffled scores of each window
    win.scr.mean <- win.scr.mu/noofperms

    ## calculate standard deviation of the shuffled scores
    ## of each window by
    ## SD(X)=SQRT(SUM(X^2)/(N-1) - (SUM(X)^2)/(N*(N-1)))
    ## while the second is an unbiased estimator using bessel's correction
    win.scr.sd <- sqrt(win.scr.sqmu/(noofperms-1) -
                       (win.scr.mu^2)/(noofperms*(noofperms-1)))

    ## calculate z-score of each original window score,
    ## depending on sign of win.scr the zscr
    ## should be upper or lower-bounded by 0,
    ## i.e., negative zscr only possible with negative window score
    win.zscr <- (win.scr - win.scr.mean)/win.scr.sd
    win.zscr <- ifelse(win.scr >= 0, pmax(0, win.zscr), pmin(0, win.zscr))

    ## output z scores
    ## NOTE: only if window length is larger than 0
    idx <- which(win.len > 0)
    ## output of zscore bed file
    if (length(idx) > 0){
      desc <- paste(win.len[idx],
                    round(TileCalcWinGC(gc=data$gc, win.idx=win.idx)[idx],
                          digits=2),
                    round(100*win.qval[idx], digits=2),
                    round(win.scr[idx], digits=2), sep="_")

      res <- c(res, list(data.frame(seq_name=data$name[idx],
         start=format(
           TileConvCenterToStart(centerpos=data$centerpos[idx],
           length=data$probelength[idx]), trim=T, scientific=F),
         end=format(
           TileConvCenterToEnd(centerpos=data$centerpos[idx + win.len[idx] - 1],
           length=data$probelength[idx + win.len[idx] - 1]),
           trim=T, scientific=F), desc=desc,
         score=round(win.zscr[idx], digits=2), strand="+",
         stringsAsFactors=FALSE)))
      rm(desc)
    }
    else {
      res <- c(res, list(data.frame()))
    }
    ## cleanup
    rm(idx, win.scr.mu, win.scr.sqmu, win.scr.mean, win.scr.sd)
  }
  else {
    res <- c(res, list(data.frame()))
  }


  ######################### JOINING CONSECUTIVE WINDOWS ########################
  if (length(win.res) > 0){
    if (verbose){
      TileMsg("Joining overlapping positive windows")
    }
    ## collect probes within highly expressed windows
    ## NOTE: not only windows with common probes will be joined but also windows
    ## that overlap in probe sequence or are directly next to each other
    ## --> same as interval union on expressed probe nucleotides
    ## POSSIBLE ADDITION: may be extended by use of an explicite gap parameter
    tmp.idx <- win.scr[win.res] >= 0
    win.red.len <- NULL
    win.red.st <- NULL
    if (any(tmp.idx)){
      tmp <- interval_union(Intervals(matrix(c(win.res[tmp.idx],
         win.res[tmp.idx] + win.len[win.res[tmp.idx]] - 1), ncol=2),
         closed=TRUE))

      if (nrow(tmp) >= 2){
        ## calculating distance between last and first probe of adjacent windows
        dist <- TileConvCenterToStart(
                  centerpos=data$centerpos[tmp@.Data[2:nrow(tmp),1]],
                  length=data$probelength[tmp@.Data[2:nrow(tmp),1]]) -
                TileConvCenterToEnd(
                  centerpos=data$centerpos[tmp@.Data[1:(nrow(tmp)-1),2]],
                  length=data$probelength[tmp@.Data[1:(nrow(tmp)-1),2]]) - 1

        ## select adjacent windows on same chromosome that overlap
        ## or are directly next to each other
        dist.idx <- which(dist <= 0 &
                          data$fullname[tmp@.Data[2:nrow(tmp),1]] ==
                          data$fullname[tmp@.Data[1:(nrow(tmp)-1),2]])

        ## merging those windows in parallel by taking the interval union
        ## of the orignal windows with Intervals overlapping the selected
        ## window boundaries
        tmp <- interval_union(tmp, Intervals(matrix(c(tmp@.Data[dist.idx,2],
           tmp@.Data[dist.idx+1,1]), ncol=2), closed=TRUE))
        rm(dist, dist.idx)
      }
      win.red.len <- tmp@.Data[,2] - tmp@.Data[,1] + 1
      win.red.st <- tmp@.Data[,1]
      rm(tmp)
    }
    rm(tmp.idx)

    ## consider negative outliers in differential expression
    if (diff){
      tmp.idx <- win.scr[win.res] < 0
      if (any(tmp.idx)){
        tmp.neg <- interval_union(Intervals(matrix(c(win.res[tmp.idx],
           win.res[tmp.idx] + win.len[win.res[tmp.idx]] - 1), ncol=2),
           closed=TRUE))
        if (nrow(tmp.neg) >= 2){
          ## calculating distance between last and first probe
          ## of adjacent windows
          dist.neg <- TileConvCenterToStart(
             centerpos=data$centerpos[tmp.neg@.Data[2:nrow(tmp.neg),1]],
             length=data$probelength[tmp.neg@.Data[2:nrow(tmp.neg),1]]) -
                      TileConvCenterToEnd(
             centerpos=data$centerpos[tmp.neg@.Data[1:(nrow(tmp.neg)-1),2]],
             length=data$probelength[tmp.neg@.Data[1:(nrow(tmp.neg)-1),2]]) - 1

          ## select adjacent windows on same chromosome that overlap or
          ## are directly next to each other
          dist.neg.idx <- which(dist.neg <= 0 &
               data$fullname[tmp.neg@.Data[2:nrow(tmp.neg),1]] ==
               data$fullname[tmp.neg@.Data[1:(nrow(tmp.neg)-1),2]])

          ## merging those windows in parallel by taking the interval union
          ## of the orignal windows with Intervals overlapping the selected
          ## window boundaries
          tmp.neg <- interval_union(tmp.neg,
             Intervals(matrix(c(tmp.neg@.Data[dist.neg.idx,2],
                                tmp.neg@.Data[dist.neg.idx+1,1]),
                              ncol=2), closed=TRUE))
          rm(dist.neg, dist.neg.idx)
        }
        win.red.len <- c(win.red.len, tmp.neg@.Data[,2] -
                         tmp.neg@.Data[,1] + 1)[order(c(win.red.st,
                                                        tmp.neg@.Data[,1]))]
        win.red.st <- c(win.red.st,
                        tmp.neg@.Data[,1])[order(c(win.red.st,
                                                   tmp.neg@.Data[,1]))]
        rm(tmp.neg)
      }
      rm(tmp.idx)
    }

    ## compute avg.score, avg.gc
    ## NOTE: median calculation may be time-expensive
    ## with large number of results and large window sizes
    if (!diff || diff.variant == "A"){
      win.red.scr <- TileCalcWinScr2(int=data$intensity, win.st=win.red.st,
                                     win.len=win.red.len,
                                     score.function=score.function)
    }
    else {
      win.red.scr <- rep(NA, length(win.red.st))
      tmp <- data$intensity
      tmp[int.minus] <- NA
      idx <- which(win.red.st %in% win.plus)
      if (length(idx) > 0){
        win.red.scr[idx] <- TileCalcWinScr2(int=tmp, win.st=win.red.st[idx],
                                            win.len=win.red.len[idx],
                                            score.function=score.function)
      }
      tmp <- data$intensity
      tmp[int.plus] <- NA
      idx <- which(win.red.st %in% win.minus)
      if (length(idx) > 0){
        win.red.scr[idx] <- TileCalcWinScr2(int=tmp, win.st=win.red.st[idx],
                                            win.len=win.red.len[idx],
                                            score.function=score.function)
      }
      stopifnot(all(!is.na(win.red.scr)))
      rm(tmp, idx)
    }
    win.red.gc <- TileCalcWinGC2(gc=data$gc, win.st=win.red.st,
                                 win.len=win.red.len)
    win.red.qval <- TileCalcWinMinQ(qval=win.qval, len=win.len,
                                    win.st=win.red.st, win.len=win.red.len)

    ## create bed dataframe
    win.red.bed <- data.frame(seq_name=data$name[win.red.st],
                              start=TileConvCenterToStart(
                                centerpos=data$centerpos[win.red.st],
                                length=data$probelength[win.red.st]),
                              end=TileConvCenterToEnd(
                                centerpos=data$centerpos[win.red.st +
                                  win.red.len - 1],
                                length=data$probelength[win.red.st +
                                  win.red.len - 1]),
                              desc=paste(data$name[win.red.st],
                                format(TileConvCenterToStart(
                                  centerpos=data$centerpos[win.red.st],
                                  length=data$probelength[win.red.st]),
                                       trim=T, scientific=F),
                                format(TileConvCenterToEnd(
                                  centerpos=data$centerpos[win.red.st +
                                    win.red.len - 1],
                                  length=data$probelength[win.red.st +
                                    win.red.len - 1]), trim=T, scientific=F),
                                format(win.red.len, trim=T, scientific=F),
                                round(win.red.gc, digits=2),
                                round(100*win.red.qval, digits=2),
                                round(win.red.scr, digits=2), sep="_"),
                              score=10, strand="+", stringsAsFactors = FALSE)
  }
  ############################# CLEANUP ########################################
  rm(win.idx, win.idx2, win.len, win.scr, win.cnt)
  if (diff){
    if (diff.variant == "A"){
      rm(win.cnt2)
    }
    else {
      rm(win.plus, win.minus, int.plus, int.minus)
      rm(win.idx.plus, win.idx.minus)
      rm(win.len2)
    }
  }
  rm(win.pval, win.qval)
  if (length(win.res) > 0){
    rm(win.red.st, win.red.len)
    rm(win.red.scr, win.red.gc, win.red.qval)
    rm(win.res)
    return(c(res, list(win.red.bed)))
  }
  else {
    return(c(res, list(data.frame())))
  }
}

################################################################################
#'
#' Checks parameter for exported function \code{\link{TileShuffle}}.
#'
#' Checks parameter for exported function \code{\link{TileShuffle}} whether
#' they are missing, have a incorrect data type, or are obviously not correct,
#' e.g., filename parameter refers to non-existing file.
#'
#' @param  input.type An \code{integer} value from the set of valid values
#'         (1, 2, or 3).
#' @param  signal.filename  Filename as a \code{character} to existing file.
#' @param  cel.filename A \code{vector} of one or more filenames as a
#'         \code{character} to existing files.
#' @param  cel2.filename A \code{vector} of one or more filename as a
#'         \code{character} to existing files or set to \code{NULL}.
#' @param  celinc.filename A \code{vector} of filenames or one or more
#'         filenames as \code{character} to existing files or set to
#'         \code{NULL}.
#' @param  custom.filename A \code{vector} of one or more filenames as a
#'         \code{character} to existing files.
#' @param  custom2.filename A \code{vector} of one or more filename as a
#'         \code{character} to existing files or set to \code{NULL}.
#' @param  custominc.filename A \code{vector} of filenames or one or more
#'         filenames as \code{character} to existing files or set to
#'         \code{NULL}.
#' @param  bpmap.filename Filename as a \code{character} to existing file.
#' @param  minhits An \code{integer} value.
#' @param  group A \code{character} value.
#' @param  pmonly A \code{logical} value.
#' @param  normalize A \code{logical} value.
#' @param  mod.tstat A \code{logical} value.
#' @param  noofperms An \code{integer} value greater than zero.
#' @param  winsize An \code{integer} value greater than zero.
#' @param  qvalue A \code{numeric} value between zero and one (both boundaries
#'         included).
#' @param  gcmode A \code{character} value from the set of valid values
#'         ("fixed", "automatic").
#' @param  gcnum An \code{integer} value. If \code{gcmode} is set to "fixed"
#'         only values from the set (1, 2, 3, 4, or 5) are valid.
#' @param  score.function A \code{character} value from the set of valid values
#'         ("mean", "trimmed", or "median").
#' @param  randomize A \code{logical} value.
#' @param  diff A \code{logical} value.
#' @param  diff.variant A \code{character} value from the set of valid values
#'         ("A" or "B"). Only required if \code{diff} is enabled.
#' @param  regions.filename Filename as a \code{character} to existing file or
#'         set to \code{NULL}.
#' @param  zscore.filename Filename as a \code{character} or set to \code{NULL}.
#' @param  output.filename Filename as a \code{character}.
#' @param  verbose A \code{logical} value.
#' @return None but the given paramters into their correct data type if
#'         necessary and possible.
#' @keywords internal
#'
TileShuffleCheck <- function(input.type,
                             signal.filename,
                             cel.filename,
                             cel2.filename,
                             celinc.filename,
                             custom.filename,
                             custom2.filename,
                             custominc.filename,
                             bpmap.filename,
                             minhits,
                             group,
                             pmonly,
                             normalize,
                             mod.tstat,
                             noofperms,
                             winsize,
                             qvalue,
                             gcmode,
                             gcnum,
                             score.function,
                             randomize,
                             diff,
                             diff.variant,
                             regions.filename,
                             zscore.filename,
                             output.filename,
                             verbose){
################################################################################
  if (missing(input.type) || !(is.numeric(input.type) ||
                               is.integer(input.type)) ||
      round(input.type) != input.type || length(input.type) != 1){
    stop("The parameter 'input.type' is missing or invalid.")
  }
  else {
    if (!(input.type %in% c(1:3))){
      stop("The parameter 'input.type' can only be set to 1, 2, or 3.")
    }
  }
  if (missing(zscore.filename) ||
      (!is.null(zscore.filename) && !is.character(zscore.filename)) ||
      (!is.null(zscore.filename) && length(zscore.filename) != 1)){
    stop("The parameter 'zscore.filename' is missing or invalid.")
  }
  if (missing(output.filename) || !is.character(output.filename) ||
      length(output.filename) != 1){
    stop("The parameter 'output.filename' is missing or invalid.")
  }
  if (input.type == 1){
    TileReadSignalCheck(signal.filename=signal.filename,
                        bpmap.filename=bpmap.filename,
                        minhits=minhits, group=group,
                        gc=TRUE, matchscore=FALSE,
                        verbose=verbose)
    
    if (!missing(cel.filename) || !is.null(cel2.filename) ||
        !is.null(celinc.filename) || !missing(custom.filename) ||
        !is.null(custom2.filename) || !is.null(custominc.filename)){
      stop("Parameters to files defined that will not be used ",
           "with the current input.type.")
    }
  }
  else if (input.type == 2){
    TileReadCelCheck(cel.filename=cel.filename, cel2.filename=cel2.filename,
                     celinc.filename=celinc.filename,
                     bpmap.filename=bpmap.filename,
                     minhits=minhits, group=group, pmonly=pmonly, mmonly=FALSE,
                     normalize=normalize, mod.tstat=mod.tstat, gc=TRUE,
                     matchscore=FALSE, verbose=verbose)

    if (!missing(signal.filename) || !missing(custom.filename) ||
        !is.null(custom2.filename) || !is.null(custominc.filename)){
      stop("Parameters to files defined that will not be used ",
           "with the current input.type.")
    }
  }
  else if (input.type == 3){
    TileReadCustomCheck(custom.filename=custom.filename,
                        custom2.filename=custom2.filename,
                        custominc.filename=custominc.filename,
                        pmonly=pmonly, mmonly=FALSE, normalize=normalize,
                        mod.tstat=mod.tstat, gc=TRUE, verbose=verbose)
    
    if (!missing(signal.filename) || !missing(bpmap.filename) ||
        !missing(cel.filename) || !is.null(cel2.filename) ||
        !is.null(celinc.filename)){
      stop("Parameters to files defined that will not be used ",
           "with the current input.type.")
    }
  }
  TileAnalysisCheck(noofperms=noofperms, winsize=winsize, qvalue=qvalue,
                    gcmode=gcmode, gcnum=gcnum, score.function=score.function,
                    randomize=randomize, diff=diff, diff.variant=diff.variant,
                    regions.filename=regions.filename,
                    zscore=!is.null(zscore.filename),
                    verbose=verbose)

  ## additional checks
  if (input.type == 2 && !is.null(cel2.filename) && diff == FALSE){
    stop("Parameter 'cel2.filename' given but no differential expression ",
         "analysis is selected (parameter 'diff' is set to FALSE)")
  }
  if (input.type == 2 && is.null(cel2.filename) && diff == TRUE){
    stop("Differential expression analysis is selected without a Affymetrix ",
         "CEL file in the second condition (parameter 'cel2.filename' not set)")
  }
}

################################################################################
#'
#' Reports the output for the exported function \code{\link{TileShuffle}}.
#'
#' Reports the output for the exported function \code{\link{TileShuffle}} as a
#' BED-formatted output file that is written and comprises information on the
#' segments identified as significantly (differentially) expressed. Moreover,
#' if \code{zscore.filename} is given accordingly (not set to \code{NULL}) an
#' addiional BED-formatted file is written that comprises the analyzed windows
#' of the statistical analysis. Note that any existing file will be overwritten.
#' Moreover, all generated files contain header information marked by `#' at the
#' beginning of the line that lists the used parameters with their values and
#' the description of the columns in the corresponding file.
#'
#' @param  bed A \code{list}, reported by \code{\link{TileAnalysis}}, with two
#'         entries where the first one stores all information on the windows
#'         including their z-score in case the \code{zscore.filename} is not set
#'         to \code{NULL} while the second one keeps the information on the
#'         segments identified as (differentially) expressed.
#' @param  input.type The type of input data which contains the probe and
#'         intensity information. In case of \code{input.type == 1}, the input
#'         is read by use of the function \code{\link{TileReadSignal}}. On the
#'         other hand, in case of \code{input.type == 2}, the input is read
#'         by using the function \code{\link{TileReadCel}} while it is read by
#'         \code{\link{TileReadCustom}} if \code{input.type == 3} is selected.
#' @param  signal.filename  Filename as a \code{character} that is required by
#'         the function \code{\link{TileReadSignal}}.
#' @param  cel.filename  A \code{vector} of one or more filenames as
#'         \code{character} that is required by the function
#'         \code{\link{TileReadCel}}.
#' @param  cel2.filename A \code{vector} of one or more filenames as
#'         \code{character} that is required by the function
#'         \code{\link{TileReadCel}} if \code{diff} is enabled.
#' @param  celinc.filename A \code{vector} of one or more filenames as
#'         \code{character} that may be provided to the function
#'         \code{\link{TileReadCel}}.
#' @param  custom.filename A \code{vector} of one or more filenames as
#'         \code{character} that is required by the function
#'         \code{\link{TileReadCustom}}.
#' @param  custom2.filename A \code{vector} of one or more filenames as
#'         \code{character} that is required by the function
#'         \code{\link{TileReadCustom}} if \code{diff} is enabled.
#' @param  custominc.filename A \code{vector} of one or more filenames as
#'         \code{character} that may be provided to the function
#'         \code{\link{TileReadCustom}}.
#' @param  bpmap.filename Filename as a \code{character} that is required by the
#'         function \code{\link{TileReadSignal}} and \code{\link{TileReadCel}}.
#' @param  minhits An \code{integer} value that is required by the function
#'         \code{\link{TileReadSignal}} and \code{\link{TileReadCel}}.
#' @param  group A \code{character} value that is required by the function
#'         \code{\link{TileReadSignal}} and \code{\link{TileReadCel}}.
#' @param  pmonly A \code{logical} value that may be provided to the function
#'         \code{\link{TileReadCel}} and \code{\link{TileReadCustom}}.
#' @param  normalize A \code{logical} value that may be provided to the function
#'         \code{\link{TileReadCel}} and \code{\link{TileReadCustom}}.
#' @param  mod.tstat A \code{logical} value that may be provided to the function
#'         \code{\link{TileReadCel}} and \code{\link{TileReadCustom}}.
#' @param  noofperms An \code{integer} value that is required by the function
#'         \code{\link{TileAnalysis}}.
#' @param  winsize An \code{integer} value that is required by the function
#'         \code{\link{TileAnalysis}}.
#' @param  qvalue A \code{numeric} value that is required by the function
#'         \code{\link{TileAnalysis}}.
#' @param  gcmode A \code{character} value that is required by the function
#'         \code{\link{TileAnalysis}}.
#' @param  gcnum An \code{integer} value that is required by the function
#'         \code{\link{TileAnalysis}}.
#' @param  score.function A \code{character} value that is required by the
#'         function \code{\link{TileAnalysis}}.
#' @param  randomize A \code{logical} value that is required by the function
#'         \code{\link{TileAnalysis}}.
#' @param  diff A \code{logical} value that is required by the function
#'         \code{\link{TileAnalysis}}.
#' @param  diff.variant A \code{character} value that is required by the
#'         function \code{\link{TileAnalysis}} if \code{diff} is enabled.
#' @param  regions.filename Filename as a \code{character} that may be provided
#'         to to the function \code{\link{TileAnalysis}}.
#' @param  zscore.filename Filename as a \code{character} where the analyzed
#'         windows by of the statistical analysis including the given
#'         information are written to (if \code{zscore.filename} is not set
#'         to \code{NULL}). Note that any existing file will be overwritten.
#' @param  output.filename Filename as a \code{character} where the identified
#'         segments including the given information are written to. Note that
#'         any existing file will be overwritten.
#' @param  verbose A \code{logical} value.
#' @return None. All output is written to the \code{output.filename} and
#'         \code{zscore.filename} if \code{zscore.filename} is not \code{NULL}.
#' @keywords internal
#'
TileShuffleOutput <- function(bed,
                              input.type,
                              signal.filename,
                              cel.filename,
                              cel2.filename,
                              celinc.filename,
                              custom.filename,
                              custom2.filename,
                              custominc.filename,
                              bpmap.filename,
                              minhits,
                              group,
                              pmonly,
                              normalize,
                              mod.tstat,
                              noofperms,
                              winsize,
                              qvalue,
                              gcmode,
                              gcnum,
                              score.function,
                              randomize,
                              diff,
                              diff.variant,
                              regions.filename,
                              zscore.filename,
                              output.filename,
                              verbose){
################################################################################
  package.name <- getPackageName()
  package.version <- packageDescription(package.name, fields = "Version")

  header <- paste("# timestamp=", date(), "\n# package=", package.name,
                  "\n# version=", package.version, "\n# input.type=",
                  input.type, sep="")
  
  if (input.type == 1){
    header <- paste(header, "\n# signal.filename=",
                    TileConvPath(filename=signal.filename),
                    "\n# bpmap.filename=",
                    TileConvPath(filename=bpmap.filename),
                    "\n# minhits=", minhits, "\n# group=", group, sep="")
  }
  if (input.type == 2){
    header <- paste(header,
                    paste("\n# cel.filename=", sep="", collapse="",
                          TileConvPath(filename=cel.filename)),
                    paste("\n# cel2.filename=", sep="", collapse="",
                          TileConvPath(filename=cel2.filename)),
                    paste("\n# celinc.filename=", sep="", collapse="",
                          TileConvPath(filename=celinc.filename)),
                    "\n# bpmap.filename=",
                    TileConvPath(filename=bpmap.filename),
                    "\n# minhits=", minhits, "\n# group=", group,
                    "\n# pmonly=", pmonly, "\n# normalize=", normalize,
                    "\n# mod.tstat=", mod.tstat, sep="")
  }
  if (input.type == 3){
      header <- paste(header,
                      paste("\n# custom.filename=", sep="", collapse="",
                            TileConvPath(filename=custom.filename)),
                      paste("\n# custom2.filename=", sep="", collapse="",
                            TileConvPath(filename=custom2.filename)),
                      paste("\n# custominc.filename=", sep="", collapse="",
                            TileConvPath(filename=custominc.filename)),
                      "\n# pmonly=", pmonly, "\n# normalize=", normalize,
                      "\n# mod.tstat=", mod.tstat, sep="")
  }

  header <- paste(header,
                  "\n# noofperms=", noofperms, "\n# winsize=", winsize,
                  "\n# qvalue=", qvalue, "\n# gcmode=", gcmode,
                  "\n# gcnum=", gcnum, "\n# score.function=", score.function,
                  "\n# randomize=", randomize, "\n# diff=", diff,
                  ifelse(diff, paste("\n# diff.variant=", diff.variant, sep=""),
                         ""), "\n# regions.filename=",
                  TileConvPath(filename=regions.filename), sep="")

  if (verbose){
    TileMsg("Writing output file ", TileConvPath(filename=output.filename))
  }

  cat(file=output.filename, header, "\n# chrname\tstart\tend\tchrname_start_",
      "end_noofprobes_avggc_minqvalue_winscore\t10\t+\n", sep="", append=FALSE)

  if (!is.null(bed[[2]]) && nrow(bed[[2]]) > 0){
    write.table(x=format(bed[[2]], scientific=F, trim=T),
                file=output.filename, quote=FALSE, sep="\t",
                row.names=FALSE, col.names=FALSE, append=TRUE)
  }

  if (!is.null(zscore.filename)){
    if (verbose){
      TileMsg("Writing zscore file ", TileConvPath(filename=zscore.filename))
    }

    cat(file=zscore.filename, header, "\n# chrname\tstart\tend\tnoofprobes_",
        "avggc_winqvalue_winscore\tzscore\t+\n", sep="", append=FALSE)

    if (!is.null(bed[[1]]) && nrow(bed[[1]]) > 0){
      write.table(format(bed[[1]], scientific=F, trim=T),
                  file=zscore.filename, quote=FALSE, sep="\t",
                  row.names=FALSE, col.names=FALSE, append=TRUE)
    }
  }
}


################################################################################
#'
#' The overall procedure to analyze tiling array expression data.
#'
#' The overall procedure to analyze tiling array expression data by reading the
#' measure probe scores and further information as input, executes the
#' statistical analysis of this data, and report the output in terms of
#' one or two BED-formatted files (depending on whether \code{zscore.filename}
#' is set to \code{NULL}).
#'
#' @param  input.type The type of input data which contains the probe and
#'         intensity information.
#'         In case of \code{input.type == 1}, the input is given in terms of
#'         human-readable signal.txt files, created by the Affymetrix Tiling
#'         Array Software (TAS) that contain the probe scores, and Affymetrix
#'         BPMAP files comprising the probe mapping and further information on
#'         them.
#'         In case of \code{input.type == 2}, the input is provided as
#'         Affymetrix CEL files and Affymetrix BPMAP files.
#'         On the other hand, by selecting \code{input.type == 3}, the input is
#'         provided as custom-formatted files that may be created from tiling
#'         array data of any design and platform. More details on the format are
#'         given in the description of the function \code{\link{TileReadCustom}}
#'         The selection of \code{input.type == 1} is deprecated since it was
#'         specifically designed for the Affymetrix pipeline with TAS on the
#'         Human tiling array 1.0R platform and may not work properly in other
#'         cases. Instead, please use \code{input.type = 2} with Affymetrix CEL
#'         files or \code{input.type = 3} with custom-formatted text files.
#' @param  signal.filename  Filename of human-readable signal.txt file
#'         (as a \code{character}) created by the Tiling Array Software (TAS).
#' @param  cel.filename A \code{vector} of one or more filenames of Affymetrix
#'         CEL files (as \code{character}) that contain the probe intensities in
#'         the first cellular condition. Note that replicates are simply defined
#'         by more than one filename and used according to \code{mod.tstat}.
#' @param  cel2.filename A \code{vector} of one or more filenames of Affymetrix
#'         CEL files (as \code{character}) that contain the probe intensities in
#'         the second cellular condition. Note that replicates are simply
#'         defined by more than one filename and used according to
#'         \code{mod.tstat}.
#' @param  celinc.filename A \code{vector} of one or more filenames of
#'         Affymetrix CEL files (as \code{character}) containing probe
#'         intensities that should be included in the normalization. This may
#'         be desirable in case tiling array data of more than two different
#'         cellular states is available and multiple transitions between them
#'         are being analyzed.
#' @param  custom.filename A \code{vector} of one or more filenames of custom-
#'         formatted files (as \code{character}) that contain the probe
#'         intensities in the first cellular condition and are formatted as
#'         described. Note that replicates are simply defined by more than
#'         one filename and used according to \code{mod.tstat}.
#' @param  custom2.filename A \code{vector} of one or more filenames of custom-
#'         formatted files (as \code{character}) that contain the probe
#'         intensities in the second cellular condition and are formatted as
#'         described. Note that replicates are simply defined by more than one
#'         one filename and used according to \code{mod.tstat}.
#' @param  custominc.filename A \code{vector} of one or more filenames of
#'         custom-formatted files (as \code{character}) containing probe
#'         intensities that should be included in the normalization. This may
#'         be desirable in case tiling array data of more than two different
#'         cellular states is available and multiple transitions between them
#'         are being analyzed.
#' @param  bpmap.filename Filename of Affymetrix binary probe mapping (BPMAP)
#'         file (as a \code{character}), which is a binary file containing
#'         information on the location of each probe in the reference sequence.
#'         Moreover, it stores the probe sequences that are necessary to
#'         calculate the GC content.
#' @param  minhits Minimal number of hits in BPMAP entry to be considered for
#'         the further analysis. Due to historical reasons there are several
#'         entries in the BPMAP file with only around thousand probes assigned
#'         that might overlap with the larger entries or with entries on other
#'         tiling arrays. In case of Affy tiling array 1.0R, a value of 8000
#'         is recommended.
#' @param  group A group name as the organism abbreviation in order to consider
#'         only these entries in the BPMAP file and hence disregard entries such
#'         as TIGR, Affymetrix, or bacterial controls.
#' @param  pmonly Indicates whether only intensities of perfect match (PM)
#'         probes on the tiling array are incorporated in the probe intensity
#'         estimation. If neither \code{pmonly} nor \code{mmonly} is set to
#'         \code{TRUE}, the specific hybridization effect of a probe is
#'         estimated by taking \code{PM-MM}.
#' @param  normalize Indicates whether the probe intensities of the given files
#'         in \code{cel.filename}, \code{cel2.filename}, and
#'         \code{celinc.filename} (with \code{input.type == 2} or
#'         \code{custom.filename}, \code{custom2.filename}, and
#'         \code{custominc.filename} (with \code{input.type == 3}) are
#'         normalized by use of full-quantile normalization. The normalization
#'         is recommended if replicates are available or a differential analysis
#'         is executed and, hence, the transition between cellular states is
#'         analyzed.
#' @param  mod.tstat Indicates the use of replicate information. If \code{TRUE},
#'         the score is the value of the moderated t-stastistic (see
#'         \code{eBayes} of limma package for further details).
#'         Otherwise, the median probe \code{log2}-intensity among the given
#'         replicates or the median of all pairwise \code{log2}-fold changes
#'         between both states will be used as estimate of the probe
#'         differential score. Note that the moderated t-statistic can only be
#'         used if replicate information is available.
#' @param  noofperms Number of permutations used to sample the background
#'         distribution. With higher number of permutations, the statistical
#'         significance of windows can be assessed more precisely in particular
#'         with more restrictive significance thresholds, i.e., low values of
#'         the \code{qvalue} parameter. Moreover, in case of the differential
#'         analysis (\code{diff} is enabled), the number of permutations should
#'         be increased to sample the background distribution more accurately.
#'         In general, \code{noofperm} is recommended to be set to \code{10000}
#'         and \code{100000} in case of expression and differential expression
#'         analyses, respectively.
#' @param  winsize Maximal width of the windows that are being statistically
#'         assessed. The width is defined as the difference in the genomic
#'         center positions of the first and last enclosed probe. Due to
#'         gaps in the commonly uniform distribution of probes over the
#'         genomic sequence. The analyzed windows may be considerably shorter
#'         than the defined \code{winsize} or may even consist of only one
#'         probe.
#' @param  qvalue Maximal permitted q-value that is applied in the statistical
#'         analysis. Hence, windows with a q-value above the given value will
#'         not be included in the returned \code{data.frame}.
#' @param  gcmode Mode of GC content binning. In case of \code{gcmode} set to
#'         "fixed", the classification of probes in bins was predefined
#'         considering the GC content effect on probe intensities on the
#'         Affymetrix tiling array 1.0R platform. In this case, only the
#'         following values are permitted: 1, 2, 3, 4, or 5. By setting
#'         \code{gcmode} to "automatic", the binning is done automatically
#'         solely on the distribution of the GC content of the probes in order
#'         to obtain GC content bins that are optimally balanced in terms of
#'         their sizes.
#' @param  gcnum Number of different GC content bins where probes within each
#'         bin have a similar expected sequence-specific affinity and are
#'         permuted independently from each other. Accordingly, intensities
#'         of probes that belong to different affinity bins must not be
#'         interchanged. Due to the trade-off between the reduction of the
#'         sequence-specific effect and the maintenance of sufficiently large
#'         permutation bins, three GC content bins are recommended in the
#'         expression analysis. In case of \code{gcmode} set to "fixed", the
#'         classification of probes in bins was predefined considering the
#'         GC content effect on probe intensities on the Affymetrix tiling
#'         array 1.0R platform. In this case, only the following values are
#'         permitted: 1, 2, 3, 4, or 5. By setting \code{gcmode} to "automatic",
#'         the binning is done automatically solely on the distribution of the
#'         GC content of the probes in order to obtain GC content bins that are
#'         optimally balanced in terms of their sizes. Note that the
#'         \code{gcnum} is set to one in case of differential expression
#'         analysis (\code{diff} enabled) since sequence-specific effect cancel
#'         out and affinity binning is rendered unnecessary.
#' @param  score.function Function to calculate windows scores over the
#'         scores of the corresponding probes, i.e., arithmetic average
#'         (\code{score.function} = "mean"), arithmetic mean trimmed by the
#'         minimal and maximal value (\code{score.function} = "trimmed"),
#'         or the median (\code{score.function} = "median"). Note that the
#'         definition of trimmed mean differs from the common one with given
#'         percentile ranges. Moreover, the resulting scores with trimmed
#'         mean may differ from the mean only in case of windows that contain
#'         more than two probes. The latter two scoring functions are
#'         recommended due to their higher robustness against outliers. However,
#'         due to the higher calculation costs, the running time increases by
#'         selecting "trimmed" or "median". Note that the function is given as
#'         \code{character}.
#' @param  randomize Indicates whether an additional permutation is applied
#'         prior to the calculation of original window scores. It is a
#'         possiblity to roughly estimate the false positive rate since under
#'         the assumption of mostly unexpressed probes no window over permuted
#'         intensities is expected to differ significantly from the background
#'         distribution.
#' @param  diff Indicates whether differential expression analysis is applied.
#' @param  diff.variant The variants of the differential expression analysis
#'         differ in score calculation, in the permutation procedure as well as
#'         in their assignment of statistical significance to windows. The
#'         \code{diff.variant} A is similar to the normal expression analysis
#'         but two-tailed p-values are estimated to regard both regulation
#'         directions, up and down. The multiple testing correction is then
#'         adjusted to account for these additional comparisons. The
#'         \code{diff.variant} B assumes that entire windows are either up-
#'         or down-regulated between conditions. The presumed direction of
#'         regulation is initially assigned to each window on the basis of its
#'         score. Subsequently, all converse probes, i.e., probes with negative
#'         score within positive windows or vice versa, are ignored and neither
#'         permuted nor incorporated into the score calculation. Consequently,
#'         positive and negative windows are compared to different background
#'         distributions. The p-value estimation and correction is done
#'         equivalent as in the case of the normal expression analyses. Both
#'         variants produce fairly similar results while the variant B is
#'         slightly superior in its performance and hence recommended.
#' @param  regions.filename Filename of BED-formatted file that contains regions
#'         to which the (differential) expression analysis should be limited to.
#'         Hence, only windows entirely enclosed in the union of these regions
#'         are statistically evaluated. Commonly, this parameter is used in
#'         order to identify highly and differentially expression (highdiff)
#'         regions by restricting the differential expression analysis
#'         (\code{diff} enabled) to regions identified as highly expressed in
#'         either one of the corresponding cellular conditions.
#' @param  zscore.filename Filename of BED-formatted z-score file that may be
#'         written and comprises the analyzed windows of the statistical
#'         analysis. Note that it is only written if \code{zscore.filename} is
#'         set accordingly. In this case, it contains the name of the reference
#'         sequence, start and end position, description, estimate z-score
#'         z-score, and `+' as strand for each analyzed window. The description
#'         is an underscore-delimitted string of the number of covered probes,
#'         the average GC content of their sequences, the window q-value
#'         multiplied by 100, and the window score calculated with the given
#'         \code{scoring.function} on the probe scores. The z-scores represent
#'         normalized window scores that are calculated by use of the sampled
#'         background. More precisely, the window z-score \code{z} is calculated
#'         by \eqn{z = \frac{x - \mu}{\sigma}} where \code{x} is the window
#'         score and \eqn{\mu} and \eqn{\sigma} are the mean and standard
#'         deviation of the permuted window scores, respectively. Note that
#'         negative z-scores indicate down-regulation and positive indicate up-
#'         regulated regions since z-scores are bounded to zero from below or
#'         from above in case of up- or down-regulation, respectively. Hence,
#'         using common expression analysis (if \code{diff} is \code{FALSE}),
#'         the z-score cannot be negative. Note that any existing file will be
#'         overwritten.
#' @param  output.filename Filename of BED-formatted output file that is written
#'         and comprises information on the segments identified as significantly
#'         (differentially) expressed including name of reference sequence,
#'         start and end position, description, score that is uniformly set to
#'         `10' (better compatibility of BED files e.g. with UCSC genome
#'         browser), and `+' as strand. The description is an underscore-
#'         delimitted string of the name of reference sequence, the start and
#'         end position, the number of probes covered by the segment, the
#'         average GC content of their sequences, the minimal q-value of windows
#'         that were merged into the segment (multiplied by 100), and the
#'         segment score calculated with the given \code{scoring.function} on
#'         the covered probes.
#' @param  verbose Indicates whether information on progress are printed.
#' @return None. All output is written to the \code{output.filename} and
#'         \code{zscore.filename} if \code{zscore.filename} is not \code{NULL}.
#'         All generated files contain header information marked by `#' at the
#'         beginning of the line that lists the used parameters with their
#'         values and the description of the columns in the subsequent output.
#'         Note that any existing file will be overwritten.
#' @include TileHelper.R
#' @include TileRead.R
#' @export
#' @examples
#' ## Note that the following example only executes if the external data
#' ## of the Starr R package is available which includes an artificial
#' ## Affymetrix BPMAP file and corresponding CEL files.
#' path <- system.file("extdata", package = "Starr")
#' if (path != ""){
#'   ## define Affymetrix BPMAP file for probe mapping
#'   bpmap.filename <- file.path(path, "Scerevisiae_tlg_chr1.bpmap")
#'   ## define Affymetrix CEL files
#'   ## here: file of control experiment (wt)
#'   wt.filename <- file.path(path, "wt_IP_chr1.cel")
#'   ## and files of real experiment with
#'   ## tagged Rpb3 with two replicates (ip)
#'   ip.filename <- c(file.path(path, "Rpb3_IP_chr1.cel"),
#'                    file.path(path, "Rpb3_IP2_chr1.cel"))
#'   stopifnot(file.exists(bpmap.filename) && all(file.exists(ip.filename)) &&
#'             file.exists(wt.filename))
#'
#'   ## identify highly expressed segments in IP state
#'   ## (only 100 permutations as example)
#'   ## Note that group is here '' (blank) for old Affy chr21/22 arrays
#'   ## but commonly it is "Hs" for Human, "Mm" for Mouse or "Dm" for Drosophila
#'   TileShuffle(bpmap.filename=bpmap.filename, cel.filename=ip.filename,
#'               input.type=2, group="", pmonly=TRUE, normalize=TRUE,
#'               noofperms=100, winsize=200, qvalue=0.05,
#'               gcnum=3, diff=FALSE, output.filename="ip_high.bed",
#'               zscore.filename="ip_high_zscore.bed", verbose=FALSE)
#'
#'   ## identify highly expressed segments in wt state
#'   ## (only 100 permutations as example)
#'   TileShuffle(bpmap.filename=bpmap.filename, cel.filename=wt.filename,
#'               input.type=2, group="", pmonly=TRUE, normalize=TRUE,
#'               noofperms=100, winsize=200, qvalue=0.05,
#'               gcnum=3, diff=FALSE, output.filename="wt_high.bed",
#'               zscore.filename="wt_high_zscore.bed", verbose=FALSE)
#'
#'   ## concatenate files
#'   file.create("ip_and_wt_high.bed")
#'   file.append("ip_and_wt_high.bed", "ip_high.bed")
#'   file.append("ip_and_wt_high.bed", "wt_high.bed")
#'
#'   ## identify highly and differentially expressed segments (highdiff)
#'   ## that are highly expressed in control or real experiment and
#'   ## significantly differentially expressed between both conditions
#'   ## (only 100 permutations as example)
#'   ## Note: common differential analysis without idenfying highdiff segments
#'   ## is executed simply by omitting the parameter 'regions.filename'
#'   TileShuffle(bpmap.filename=bpmap.filename, cel.filename=wt.filename,
#'               cel2.filename=ip.filename, input.type=2, group="",
#'               pmonly=TRUE, normalize=TRUE, noofperms=100,
#'               winsize=200,qvalue=0.05, gcnum=1, diff=TRUE,
#'               regions.filename="ip_and_wt_high.bed",
#'               output.filename="ip_and_wt_highdiff.bed",
#'               zscore.filename="ip_and_wt_highdiff_zscore.bed",
#'               verbose=FALSE)
#'   ## cleanup
#'   rm(bpmap.filename, wt.filename, ip.filename)
#' }
#' rm(path)
#'
TileShuffle <- function(input.type,
                        signal.filename,
                        cel.filename,
                        cel2.filename = NULL,
                        celinc.filename = NULL,
                        custom.filename,
                        custom2.filename = NULL,
                        custominc.filename = NULL,
                        bpmap.filename,
                        minhits = 8000,
                        group = "Hs",
                        pmonly = TRUE,
                        normalize = TRUE,
                        mod.tstat = FALSE,
                        noofperms,
                        winsize,
                        qvalue,
                        gcmode = "fixed",
                        gcnum,
                        score.function = "trimmed",
                        randomize = FALSE,
                        diff,
                        diff.variant = "B",
                        regions.filename = NULL,
                        zscore.filename = NULL,
                        output.filename,
                        verbose = FALSE){
################################################################################

  ## parameter checks
  TileShuffleCheck(input.type=input.type, signal.filename=signal.filename,
                   cel.filename=cel.filename, cel2.filename=cel2.filename,
                   celinc.filename=celinc.filename,
                   custom.filename=custom.filename,
                   custom2.filename=custom2.filename,
                   custominc.filename=custominc.filename,
                   bpmap.filename=bpmap.filename,
                   minhits=minhits, group=group, pmonly=pmonly,
                   normalize=normalize, mod.tstat=mod.tstat,
                   noofperms=noofperms, winsize=winsize,
                   qvalue=qvalue, gcmode=gcmode, gcnum=gcnum,
                   score.function=score.function, randomize=randomize,
                   diff=diff, diff.variant=diff.variant,
                   regions.filename=regions.filename,
                   zscore.filename=zscore.filename,
                   output.filename=output.filename,
                   verbose=verbose)

  ## reading input
  if (input.type == 1){
    data <- TileReadSignal(signal.filename=signal.filename,
                           bpmap.filename=bpmap.filename,
                           minhits=minhits, group=group, gc=TRUE,
                           matchscore=FALSE, verbose=verbose)
  }
  else if (input.type == 2){
    data <- TileReadCel(cel.filename=cel.filename,
                        cel2.filename=cel2.filename,
                        celinc.filename=celinc.filename,
                        bpmap.filename=bpmap.filename,
                        minhits=minhits, group=group,
                        pmonly=pmonly, mmonly=FALSE,
                        normalize=normalize, mod.tstat=mod.tstat,
                        gc=TRUE, matchscore=FALSE, verbose=verbose)
  }
  else if (input.type == 3){
      data <- TileReadCustom(custom.filename=custom.filename,
                             custom2.filename=custom2.filename,
                             custominc.filename=custominc.filename,
                             pmonly=pmonly, mmonly=FALSE,
                             normalize=normalize, mod.tstat=mod.tstat,
                             gc=TRUE, verbose=verbose)
  }

  if (is.null(data) || nrow(data) == 0){
    stop("Reading input has failed.")
  }

  ## processing
  bed <- TileAnalysis(data=data, noofperms=noofperms, winsize=winsize,
                      qvalue=qvalue, gcmode=gcmode, gcnum=gcnum,
                      score.function=score.function, randomize=randomize,
                      diff=diff, diff.variant=diff.variant,
                      regions.filename=regions.filename,
                      zscore=!is.null(zscore.filename),
                      verbose=verbose)

  if (is.null(bed) || length(bed) != 2){
    stop("Analyzing data has failed.")
  }

  ## output
  TileShuffleOutput(bed=bed, input.type=input.type,
                    signal.filename=signal.filename, cel.filename=cel.filename,
                    cel2.filename=cel2.filename,
                    celinc.filename=celinc.filename,
                    custom.filename=custom.filename,
                    custom2.filename=custom2.filename,
                    custominc.filename=custominc.filename,
                    bpmap.filename=bpmap.filename, minhits=minhits, group=group,
                    pmonly=pmonly, normalize=normalize, mod.tstat=mod.tstat,
                    noofperms=noofperms, winsize=winsize, qvalue=qvalue,
                    gcmode=gcmode, gcnum=gcnum, score.function=score.function,
                    randomize=randomize, diff=diff, diff.variant=diff.variant,
                    regions.filename=regions.filename,
                    zscore.filename=zscore.filename,
                    output.filename=output.filename,
                    verbose=verbose)
  ## cleanup
  rm(data, bed)
}
